#!/usr/bin/python3
# ----------------------------------------------------------------------
#    Copyright (C) 2014 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import unittest

import apparmor.severity as severity
from apparmor.common import AppArmorBug, AppArmorException, cmd, hasher
from apparmor.logparser import ReadLog
from apparmor.rule.capability import CapabilityRule, CapabilityRuleset, capability_keywords
from apparmor.translations import init_translation
from common_test import AATest, setup_all_loops

_ = init_translation()


# --- check if the keyword list is up to date --- #

class CapabilityKeywordsTest(AATest):
    def test_capability_keyword_list(self):
        rc, output = cmd('../../common/list_capabilities.sh')
        self.assertEqual(rc, 0)

        cap_list = output.replace('CAP_', '').strip().lower().split('\n')

        missing_caps = []
        for keyword in cap_list:
            if keyword not in capability_keywords:
                # keywords missing in the system are ok (= older kernel), but cap_list needs to have the full list
                missing_caps.append(keyword)

        self.assertEqual(
            missing_caps, [],
            'Missing capabilities in CapabilityRule capabilities list. This test is likely running '
            'on an newer kernel and will require updating the list of capability keywords in '
            'utils/apparmor/rule/capability.py')


# --- tests for single CapabilityRule --- #
class CapabilityTest(AATest):
    def _compare_obj_with_rawrule(self, rawrule, expected):

        obj = CapabilityRule.create_instance(rawrule)

        self.assertTrue(CapabilityRule.match(rawrule))
        self.assertEqual(rawrule.strip(), obj.raw_rule)

        self._compare_obj(obj, expected)

    def _compare_obj(self, obj, expected):
        self.assertEqual(expected['allow_keyword'], obj.allow_keyword)
        self.assertEqual(expected['audit'], obj.audit)
        self.assertEqual(expected['capability'], obj.capability)
        self.assertEqual(expected['all_caps'], obj.all_caps)
        self.assertEqual(expected['deny'], obj.deny)
        self.assertEqual(expected['comment'], obj.comment)

    def test_cap_allow_all(self):
        self._compare_obj_with_rawrule("capability,", {
            'allow_keyword': False,
            'deny': False,
            'audit': False,
            'capability': set(),
            'all_caps': True,
            'comment': "",
        })

    def test_cap_allow_sys_admin(self):
        self._compare_obj_with_rawrule("capability sys_admin,", {
            'allow_keyword': False,
            'deny': False,
            'audit': False,
            'capability': {'sys_admin'},
            'all_caps': False,
            'comment': "",
        })

    def test_cap_deny_sys_admin(self):
        self._compare_obj_with_rawrule("     deny capability sys_admin,  # some comment", {
            'allow_keyword': False,
            'deny': True,
            'audit': False,
            'capability': {'sys_admin'},
            'all_caps': False,
            'comment': " # some comment",
        })

    def test_cap_multi(self):
        self._compare_obj_with_rawrule("capability sys_admin dac_override,", {
            'allow_keyword': False,
            'deny': False,
            'audit': False,
            'capability': {'sys_admin', 'dac_override'},
            'all_caps': False,
            'comment': "",
        })

    # Template for test_cap_* functions
    # def test_cap_(self):
    #     self._compare_obj_with_rawrule("capability,", {
    #         'allow_keyword': False,
    #         'deny': False,
    #         'audit': False,
    #         'capability': set(),  # (or {'foo'} if not empty)
    #         'all_caps': False,
    #         'comment': "",
    #     })

    def test_cap_from_log(self):
        parser = ReadLog('', '', '')
        event = 'type=AVC msg=audit(1415403814.628:662): apparmor="ALLOWED" operation="capable" profile="/bin/ping" pid=15454 comm="ping" capability=13  capname="net_raw"'

        parsed_event = parser.parse_event(event)

        self.assertEqual(parsed_event, {
            'request_mask': None,
            'denied_mask': None,
            'error_code': 0,
            'magic_token': 0,
            'parent': 0,
            'profile': '/bin/ping',
            'operation': 'capable',
            'resource': None,
            'info': None,
            'aamode': 'PERMITTING',
            'time': 1415403814,
            'active_hat': None,
            'pid': 15454,
            'task': 0,
            'comm': 'ping',
            'attr': None,
            'name2': None,
            'name': 'net_raw',
            'family': None,
            'protocol': None,
            'sock_type': None,
            'class': None,
        })

        obj = CapabilityRule(parsed_event['name'], log_event=parsed_event)

        self._compare_obj(obj, {
            'allow_keyword': False,
            'deny': False,
            'audit': False,
            'capability': {'net_raw'},
            'all_caps': False,
            'comment': "",
        })

        self.assertEqual(obj.get_raw(1), '  capability net_raw,')

    # def test_cap_from_invalid_log(self):
    #     parser = ReadLog('', '', '')
    #     # invalid log entry, name= should contain the capability name
    #     event = 'type=AVC msg=audit(1415403814.628:662): apparmor="ALLOWED" operation="capable" profile="/bin/ping" pid=15454 comm="ping" capability=13  capname=""'
    #
    #     parsed_event = parser.parse_event(event)
    #
    #     obj = CapabilityRule()
    #
    #     with self.assertRaises(AppArmorBug):
    #         obj.set_log(parsed_event)
    #
    #     with self.assertRaises(AppArmorBug):
    #         obj.get_raw(1)
    #
    # def test_cap_from_non_cap_log(self):
    #     parser = ReadLog('', '', '')
    #     # log entry for different rule type
    #     event = 'type=AVC msg=audit(1415403814.973:667): apparmor="ALLOWED" operation="setsockopt" profile="/home/sys-tmp/ping" pid=15454 comm="ping" lport=1 family="inet" sock_type="raw" protocol=1'
    #
    #     parsed_event = parser.parse_event(event)
    #
    #     obj = CapabilityRule()
    #
    #     with self.assertRaises(AppArmorBug):
    #         obj.set_log(parsed_event)
    #
    #     with self.assertRaises(AppArmorBug):
    #         obj.get_raw(1)

    def test_cap_from_init_01(self):
        obj = CapabilityRule('chown')

        self._compare_obj(obj, {
            'allow_keyword': False,
            'deny': False,
            'audit': False,
            'capability': {'chown'},
            'all_caps': False,
            'comment': "",
        })

    def test_cap_from_init_02(self):
        obj = CapabilityRule(['chown'])

        self._compare_obj(obj, {
            'allow_keyword': False,
            'deny': False,
            'audit': False,
            'capability': {'chown'},
            'all_caps': False,
            'comment': "",
        })

    def test_cap_from_init_03(self):
        obj = CapabilityRule('chown', audit=True, deny=True)

        self._compare_obj(obj, {
            'allow_keyword': False,
            'deny': True,
            'audit': True,
            'capability': {'chown'},
            'all_caps': False,
            'comment': "",
        })

    def test_cap_from_init_04(self):
        obj = CapabilityRule(['chown', 'fsetid'], deny=True)

        self._compare_obj(obj, {
            'allow_keyword': False,
            'deny': True,
            'audit': False,
            'capability': {'chown', 'fsetid'},
            'all_caps': False,
            'comment': "",
        })


class InvalidCapabilityTest(AATest):
    def _check_invalid_rawrule(self, rawrule):
        obj = None
        with self.assertRaises(AppArmorException):
            obj = CapabilityRule.create_instance(rawrule)

        self.assertFalse(CapabilityRule.match(rawrule))
        self.assertIsNone(obj, 'CapbilityRule handed back an object unexpectedly')

    def test_invalid_cap_missing_comma(self):
        self._check_invalid_rawrule('capability')  # missing comma

    def test_invalid_cap_non_CapabilityRule(self):
        self._check_invalid_rawrule('network,')  # not a capability rule

    def test_empty_cap_set(self):
        obj = CapabilityRule('chown')
        obj.capability.clear()
        # no capability set, and ALL not set
        with self.assertRaises(AppArmorBug):
            obj.get_clean(1)

    def test_empty_cap_list(self):
        with self.assertRaises(AppArmorBug):
            CapabilityRule([])

    def test_no_cap_list_arg(self):
        with self.assertRaises(TypeError):
            CapabilityRule()

    def test_space_cap(self):
        with self.assertRaises(AppArmorBug):
            CapabilityRule('    ')  # the whitespace capability ;-)

    def test_space_list_1(self):
        with self.assertRaises(AppArmorBug):
            CapabilityRule(['    ', '   ', '   '])  # the whitespace capability ;-)

    def test_space_list_2(self):
        with self.assertRaises(AppArmorBug):
            CapabilityRule(['chown', '   ', 'setgid'])  # includes the whitespace capability ;-)

    def test_wrong_type_for_cap(self):
        with self.assertRaises(AppArmorBug):
            CapabilityRule(dict())


class WriteCapabilityTest(AATest):
    def _check_write_rule(self, rawrule, cleanrule):
        obj = CapabilityRule.create_instance(rawrule)
        clean = obj.get_clean()
        raw = obj.get_raw()

        self.assertTrue(CapabilityRule.match(rawrule))
        self.assertEqual(cleanrule.strip(), clean, 'unexpected clean rule')
        self.assertEqual(rawrule.strip(), raw, 'unexpected raw rule')

    def test_write_all(self):
        self._check_write_rule('     capability      ,    # foo        ', 'capability, # foo')

    def test_write_sys_admin(self):
        self._check_write_rule('    audit     capability sys_admin,', 'audit capability sys_admin,')

    def test_write_sys_multi(self):
        self._check_write_rule('   deny capability      sys_admin      audit_write,# foo bar', 'deny capability audit_write sys_admin, # foo bar')

    def test_write_manually(self):
        obj = CapabilityRule(['sys_ptrace', 'audit_write'], allow_keyword=True)

        expected = '    allow capability audit_write sys_ptrace,'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')

    def test_write_priority_1(self):
        self._check_write_rule('    priority  = 923   audit     capability sys_admin,', 'priority=923 audit capability sys_admin,')

    def test_write_priority_2(self):
        self._check_write_rule('    priority  = 0     audit     capability sys_admin,', 'priority=0 audit capability sys_admin,')

    def test_write_priority_3(self):
        self._check_write_rule('    priority=-12      audit     capability sys_admin,', 'priority=-12 audit capability sys_admin,')

    def test_write_priority_4(self):
        self._check_write_rule('    priority=+99      audit     capability sys_admin,', 'priority=99 audit capability sys_admin,')


class CapabilityCoveredTest(AATest):
    def _is_covered(self, obj, rule_to_test):
        self.assertTrue(CapabilityRule.match(rule_to_test))
        return obj.is_covered(CapabilityRule.create_instance(rule_to_test))

    def _is_covered_exact(self, obj, rule_to_test):
        self.assertTrue(CapabilityRule.match(rule_to_test))
        return obj.is_covered(CapabilityRule.create_instance(rule_to_test), True, True)

    def _is_equal(self, obj, rule_to_test, strict):
        self.assertTrue(CapabilityRule.match(rule_to_test))
        return obj.is_equal(CapabilityRule.create_instance(rule_to_test), strict)

    def test_covered_single(self):
        obj = CapabilityRule.create_instance('capability sys_admin,')

        self.assertTrue(self._is_covered(obj, 'capability sys_admin,'))

        self.assertFalse(self._is_covered(obj, 'audit capability sys_admin,'))
        self.assertFalse(self._is_covered(obj, 'audit capability,'))
        self.assertFalse(self._is_covered(obj, 'capability chown,'))
        self.assertFalse(self._is_covered(obj, 'capability,'))

    def test_covered_audit(self):
        obj = CapabilityRule.create_instance('audit capability sys_admin,')

        self.assertTrue(self._is_covered(obj, 'capability sys_admin,'))
        self.assertTrue(self._is_covered(obj, 'audit capability sys_admin,'))

        self.assertFalse(self._is_covered(obj, 'audit capability,'))
        self.assertFalse(self._is_covered(obj, 'capability chown,'))
        self.assertFalse(self._is_covered(obj, 'capability,'))

    def test_covered_check_audit(self):
        obj = CapabilityRule.create_instance('audit capability sys_admin,')

        self.assertFalse(self._is_covered_exact(obj, 'capability sys_admin,'))
        self.assertTrue(self._is_covered_exact(obj, 'audit capability sys_admin,'))

        self.assertFalse(self._is_covered_exact(obj, 'audit capability,'))
        self.assertFalse(self._is_covered_exact(obj, 'capability chown,'))
        self.assertFalse(self._is_covered_exact(obj, 'capability,'))

    def test_equal(self):
        obj = CapabilityRule.create_instance('capability sys_admin,')

        self.assertTrue(self._is_equal(obj, 'capability sys_admin,', True))
        self.assertFalse(self._is_equal(obj, 'allow capability sys_admin,', True))
        self.assertFalse(self._is_equal(obj, 'allow capability sys_admin,', True))
        self.assertFalse(self._is_equal(obj, 'audit capability sys_admin,', True))

        self.assertTrue(self._is_equal(obj, 'capability sys_admin,', False))
        self.assertTrue(self._is_equal(obj, 'allow capability sys_admin,', False))
        self.assertFalse(self._is_equal(obj, 'audit capability sys_admin,', False))

    def test_covered_multi(self):
        obj = CapabilityRule.create_instance('capability audit_write sys_admin,')

        self.assertTrue(self._is_covered(obj, 'capability sys_admin,'))
        self.assertTrue(self._is_covered(obj, 'capability audit_write,'))
        self.assertTrue(self._is_covered(obj, 'capability audit_write sys_admin,'))
        self.assertTrue(self._is_covered(obj, 'capability sys_admin audit_write,'))

        self.assertFalse(self._is_covered(obj, 'audit capability,'))
        self.assertFalse(self._is_covered(obj, 'capability chown,'))
        self.assertFalse(self._is_covered(obj, 'capability,'))

    def test_covered_all(self):
        obj = CapabilityRule.create_instance('capability,')

        self.assertTrue(self._is_covered(obj, 'capability sys_admin,'))
        self.assertTrue(self._is_covered(obj, 'capability audit_write,'))
        self.assertTrue(self._is_covered(obj, 'capability audit_write sys_admin,'))
        self.assertTrue(self._is_covered(obj, 'capability sys_admin audit_write,'))
        self.assertTrue(self._is_covered(obj, 'capability,'))

        self.assertFalse(self._is_covered(obj, 'audit capability,'))

    def test_covered_deny(self):
        obj = CapabilityRule.create_instance('capability sys_admin,')

        self.assertTrue(self._is_covered(obj, 'capability sys_admin,'))

        self.assertFalse(self._is_covered(obj, 'audit deny capability sys_admin,'))
        self.assertFalse(self._is_covered(obj, 'deny capability sys_admin,'))
        self.assertFalse(self._is_covered(obj, 'capability chown,'))
        self.assertFalse(self._is_covered(obj, 'capability,'))

    def test_covered_deny_2(self):
        obj = CapabilityRule.create_instance('deny capability sys_admin,')

        self.assertTrue(self._is_covered(obj, 'deny capability sys_admin,'))

        self.assertFalse(self._is_covered(obj, 'audit deny capability sys_admin,'))
        self.assertFalse(self._is_covered(obj, 'capability sys_admin,'))
        self.assertFalse(self._is_covered(obj, 'deny capability chown,'))
        self.assertFalse(self._is_covered(obj, 'deny capability,'))

    def test_invalid_is_covered(self):
        raw_rule = 'capability sys_admin,'

        class SomeOtherClass(CapabilityRule):
            pass

        obj = CapabilityRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_borked_obj_is_covered(self):
        obj = CapabilityRule.create_instance('capability sys_admin,')

        testobj = CapabilityRule('chown')
        testobj.capability.clear()

        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_invalid_is_equal(self):
        raw_rule = 'capability sys_admin,'

        class SomeOtherClass(CapabilityRule):
            pass

        obj = CapabilityRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_equal(testobj)

    def test_empty_init(self):
        # add to internal set instead of using .set_* (which overwrites the internal set) to make sure obj and obj2 use separate storage
        obj = CapabilityRule('fsetid')
        obj2 = CapabilityRule('fsetid')
        obj.capability.add('sys_admin')
        obj2.capability.add('sys_ptrace')

        self.assertTrue(self._is_covered(obj, 'capability sys_admin,'))
        self.assertFalse(self._is_covered(obj, 'capability sys_ptrace,'))
        self.assertFalse(self._is_covered(obj2, 'capability sys_admin,'))
        self.assertTrue(self._is_covered(obj2, 'capability sys_ptrace,'))


class CapabiliySeverityTest(AATest):
    tests = (
        ('fsetid',                      9),
        ('dac_read_search',             7),
        (['fsetid', 'dac_read_search'], 9),
        (CapabilityRule.ALL,            10),
    )

    def _run_test(self, params, expected):
        sev_db = severity.Severity('../severity.db', 'unknown')
        obj = CapabilityRule(params)
        rank = obj.severity(sev_db)
        self.assertEqual(rank, expected)

    def test_all_caps(self):
        ''' make sure all capabilities have a severity defined '''

        sev_db = severity.Severity('../severity.db', 'unknown')

        for cap in capability_keywords:
            obj = CapabilityRule(cap)
            rank = obj.severity(sev_db)
            # capabilities have a severity of 7..10, with the exception of 0 for the unused CAP_NET_BROADCAST
            # (might need adjustment if a new capability gets a different severity assigned)
            self.assertTrue(rank in [0, 7, 8, 9, 10], 'unexpected severity for capability %s: %s' % (cap, rank))

    def test_unknown_cap(self):
        sev_db = severity.Severity('../severity.db', 'unknown')
        obj = CapabilityRule('sys_admin')
        obj.capability = {'unknown_and_broken'}  # override capability with an unknown one to test for 'unknown' severity (creating obj with this invalid capability would raise an error)
        rank = obj.severity(sev_db)
        self.assertEqual(rank, 'unknown')


class CapabilityLogprofHeaderTest(AATest):
    tests = (
        ('capability,',                    [                              _('Capability'), _('ALL')]),  # noqa: E201
        ('capability chown,',              [                              _('Capability'), 'chown']),  # noqa: E201
        ('capability chown fsetid,',       [                              _('Capability'), 'chown fsetid']),  # noqa: E201
        ('audit capability,',              [_('Qualifier'), 'audit',      _('Capability'), _('ALL')]),
        ('deny capability chown,',         [_('Qualifier'), 'deny',       _('Capability'), 'chown']),
        ('allow capability chown fsetid,', [_('Qualifier'), 'allow',      _('Capability'), 'chown fsetid']),
        ('audit deny capability,',         [_('Qualifier'), 'audit deny', _('Capability'), _('ALL')]),
    )

    def _run_test(self, params, expected):
        obj = CapabilityRule.create_instance(params)
        self.assertEqual(obj.logprof_header(), expected)


# --- tests for CapabilityRuleset --- #
class CapabilityRulesTest(AATest):
    def test_empty_ruleset(self):
        ruleset = CapabilityRuleset()
        ruleset_2 = CapabilityRuleset()
        self.assertEqual([], ruleset.get_raw(2))
        self.assertEqual([], ruleset.get_clean(2))
        self.assertEqual([], ruleset_2.get_raw(2))
        self.assertEqual([], ruleset_2.get_clean(2))

    def test_ruleset_1(self):
        ruleset = CapabilityRuleset()
        rules = [
            'capability sys_admin,',
            'capability chown,',
        ]

        expected_raw = [
            'capability sys_admin,',
            'capability chown,',
            '',
        ]

        expected_clean = [
            'capability chown,',
            'capability sys_admin,',
            '',
        ]

        for rule in rules:
            ruleset.add(CapabilityRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw())
        self.assertEqual(expected_clean, ruleset.get_clean())

    def test_ruleset_2(self):
        ruleset = CapabilityRuleset()
        rules = [
            'capability chown,',
            'allow capability sys_admin,',
            'deny capability fowner, # example comment',
        ]

        expected_raw = [
            '  capability chown,',
            '  allow capability sys_admin,',
            '  deny capability fowner, # example comment',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  allow capability sys_admin,',
            '  capability chown,',
            '',
        ]

        for rule in rules:
            ruleset.add(CapabilityRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw(1))
        self.assertEqual(expected_clean, ruleset.get_clean(1))

    def test_ruleset_add(self):
        rule = CapabilityRule('fowner', comment=' # example comment')

        ruleset = CapabilityRuleset()
        ruleset.add(rule)

        expected_raw = [
            '  capability fowner, # example comment',
            '',
        ]

        expected_clean = expected_raw

        self.assertEqual(expected_raw, ruleset.get_raw(1))
        self.assertEqual(expected_clean, ruleset.get_clean(1))


class CapabilityRulesCoveredTest(AATest):
    def AASetup(self):
        self.ruleset = CapabilityRuleset()
        rules = [
            'capability chown,',
            'capability setuid setgid,',
            'allow capability sys_admin,',
            'audit capability kill,',
            'deny capability fowner, # example comment',
        ]

        for rule in rules:
            self.ruleset.add(CapabilityRule.create_instance(rule))

    def test_ruleset_is_covered_1(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('capability chown,')))

    def test_ruleset_is_covered_2(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('capability sys_admin,')))

    def test_ruleset_is_covered_3(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('allow capability sys_admin,')))

    def test_ruleset_is_covered_4(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('capability setuid,')))

    def test_ruleset_is_covered_5(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('allow capability setgid,')))

    def test_ruleset_is_covered_6(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('capability setgid setuid,')))

    def test_ruleset_is_covered_7(self):
        pass  # self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('capability sys_admin chown,')))  # fails because it is split over two rule objects internally

    def test_ruleset_is_covered_8(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('capability kill,')))

    # deny
    def test_ruleset_is_covered_9(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('deny capability chown,')))

    def test_ruleset_is_covered_10(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('deny capability sys_admin,')))

    def test_ruleset_is_covered_11(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('deny capability sys_admin chown,')))

    def test_ruleset_is_covered_12(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('deny capability setgid,')))

    def test_ruleset_is_covered_13(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('deny capability kill,')))

    # audit
    def test_ruleset_is_covered_14(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('audit capability chown,')))

    def test_ruleset_is_covered_15(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('audit capability sys_admin,')))

    def test_ruleset_is_covered_16(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('audit capability sys_admin chown,')))

    def test_ruleset_is_covered_17(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('audit capability setgid,')))

    def test_ruleset_is_covered_18(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('audit capability kill,')))

    # combined flags
    def test_ruleset_is_covered_19(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('deny capability fowner,')))

    def test_ruleset_is_covered_20(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('audit deny capability fowner,')))

    def test_ruleset_is_covered_21(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('audit capability fowner,')))

    def test_ruleset_is_covered_22(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('capability fowner,')))

    def test_ruleset_is_covered_23(self):
        self.assertTrue(self.ruleset.is_covered(CapabilityRule.create_instance('capability fowner,'), check_allow_deny=False))

    def test_ruleset_is_covered_24(self):
        self.assertFalse(self.ruleset.is_covered(CapabilityRule.create_instance('deny capability chown,'), check_allow_deny=False))

# XXX - disabling these until we decide whether or not checking whether
# a log is covered by rules should be a separate entry point, possibly
# handling the log structure directly, or whether coverage should be
# solely based on Rule objects and marshaling of a log message into a
# Rule object should occur outside of the Rule classes themselves.
#
#    def _test_log_covered(self, expected, capability):
#        event_base = 'type=AVC msg=audit(1415403814.628:662): apparmor="ALLOWED" operation="capable" profile="/bin/ping" pid=15454 comm="ping" capability=13  capname="%s"'

#        parser = ReadLog('', '', '')
#        self.assertEqual(expected, self.ruleset.is_log_covered(parser.parse_event(event_base%capability)))
#
#    def test_ruleset_is_log_covered_1(self):
#        self._test_log_covered(False, 'net_raw')
#    def test_ruleset_is_log_covered_2(self):
#        self._test_log_covered(True, 'chown')
#    def test_ruleset_is_log_covered_3(self):
#        self._test_log_covered(True, 'sys_admin')
#    def test_ruleset_is_log_covered_4(self):
#        self._test_log_covered(True, 'kill')
#    def test_ruleset_is_log_covered_5(self):
#        self._test_log_covered(False, 'fowner')
#    def test_ruleset_is_log_covered_6(self):
#        event_base = 'type=AVC msg=audit(1415403814.628:662): apparmor="ALLOWED" operation="capable" profile="/bin/ping" pid=15454 comm="ping" capability=13  capname="%s"'
#
#        parser = ReadLog('', '', '')
#        self.assertEqual(True, self.ruleset.is_log_covered(parser.parse_event(event_base%'fowner'), False))  # ignores allow/deny


class CapabilityGlobTest(AATest):
    def AASetup(self):
        self.ruleset = CapabilityRuleset()

    def test_glob(self):
        self.assertEqual(self.ruleset.get_glob('capability net_raw,'), 'capability,')

    def test_glob_ext(self):
        with self.assertRaises(NotImplementedError):
            self.ruleset.get_glob_ext('capability net_raw,')


class CapabilityDeleteTest(AATest):
    def AASetup(self):
        self.ruleset = CapabilityRuleset()
        rules = [
            'capability chown,',
            'allow capability sys_admin,',
            'deny capability fowner, # example comment',
        ]

        for rule in rules:
            self.ruleset.add(CapabilityRule.create_instance(rule))

    def test_delete(self):
        expected_raw = [
            '  capability chown,',
            '  deny capability fowner, # example comment',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  capability chown,',
            '',
        ]

        self.ruleset.delete(CapabilityRule(['sys_admin']))

        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_with_allcaps(self):
        expected_raw = [
            '  capability chown,',
            '  deny capability fowner, # example comment',
            '  capability,',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  capability chown,',
            '  capability,',
            '',
        ]

        self.ruleset.add(CapabilityRule(CapabilityRule.ALL))
        self.ruleset.delete(CapabilityRule('sys_admin'))

        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_with_multi(self):
        expected_raw = [
            '  capability chown,',
            '  allow capability sys_admin,',
            '  deny capability fowner, # example comment',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  allow capability sys_admin,',
            '  capability chown,',
            '',
        ]

        self.ruleset.add(CapabilityRule(['audit_read', 'audit_write']))
        self.ruleset.delete(CapabilityRule(['audit_read', 'audit_write']))

        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_with_multi_2(self):
        self.ruleset.add(CapabilityRule(['audit_read', 'audit_write']))

        with self.assertRaises(AppArmorBug):
            # XXX ideally delete_raw should remove audit_read from the "capability audit_read audit_write," ruleset
            #     but that's quite some work to cover a corner case.
            self.ruleset.delete(CapabilityRule('audit_read'))

    def test_delete_raw_notfound(self):
        with self.assertRaises(AppArmorBug):
            self.ruleset.delete(CapabilityRule('audit_write'))

    def test_delete_duplicates(self):
        inc = CapabilityRuleset()
        rules = [
            'capability chown,',
            'deny capability fowner, # example comment',
        ]

        for rule in rules:
            inc.add(CapabilityRule.create_instance(rule))

        expected_raw = [
            '  allow capability sys_admin,',
            '',
        ]

        expected_clean = expected_raw

        self.assertEqual(self.ruleset.delete_duplicates(inc), 2)
        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_duplicates_2(self):
        inc = CapabilityRuleset()
        rules = [
            'capability audit_write,',
            'capability fowner, # example comment',
        ]

        for rule in rules:
            inc.add(CapabilityRule.create_instance(rule))

        expected_raw = [
            '  capability chown,',
            '  allow capability sys_admin,',
            '  deny capability fowner, # example comment',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  allow capability sys_admin,',
            '  capability chown,',
            '',
        ]

        self.assertEqual(self.ruleset.delete_duplicates(inc), 0)
        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_duplicates_3(self):
        self.ruleset.add(CapabilityRule.create_instance('audit capability dac_override,'))

        inc = CapabilityRuleset()
        rules = [
            'capability dac_override,',
        ]

        for rule in rules:
            inc.add(CapabilityRule.create_instance(rule))

        expected_raw = [
            '  capability chown,',
            '  allow capability sys_admin,',
            '  deny capability fowner, # example comment',
            '  audit capability dac_override,',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  allow capability sys_admin,',
            '  audit capability dac_override,',
            '  capability chown,',
            '',
        ]

        self.assertEqual(self.ruleset.delete_duplicates(inc), 0)
        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_duplicates_4(self):
        inc = CapabilityRuleset()
        rules = ['capability,']

        for rule in rules:
            inc.add(CapabilityRule.create_instance(rule))

        expected_raw = [
            '  deny capability fowner, # example comment',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
        ]

        self.assertEqual(self.ruleset.delete_duplicates(inc), 2)
        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_duplicates_none(self):
        expected_raw = [
            '  capability chown,',
            '  allow capability sys_admin,',
            '  deny capability fowner, # example comment',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  allow capability sys_admin,',
            '  capability chown,',
            '',
        ]

        self.assertEqual(self.ruleset.delete_duplicates(None), 0)
        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def test_delete_duplicates_hasher(self):
        expected_raw = [
            '  capability chown,',
            '  allow capability sys_admin,',
            '  deny capability fowner, # example comment',
            '',
        ]

        expected_clean = [
            '  deny capability fowner, # example comment',
            '',
            '  allow capability sys_admin,',
            '  capability chown,',
            '',
        ]

        self.assertEqual(self.ruleset.delete_duplicates(hasher()), 0)
        self.assertEqual(expected_raw, self.ruleset.get_raw(1))
        self.assertEqual(expected_clean, self.ruleset.get_clean(1))

    def _check_test_delete_duplicates_in_profile(self, rules, expected_raw, expected_clean, expected_deleted):
        obj = CapabilityRuleset()

        for rule in rules:
            obj.add(CapabilityRule.create_instance(rule))

        deleted = obj.delete_duplicates(None)

        self.assertEqual(expected_raw, obj.get_raw(1))
        self.assertEqual(expected_clean, obj.get_clean(1))
        self.assertEqual(deleted, expected_deleted)

    def test_delete_duplicates_in_profile_01(self):
        rules = [
            'audit capability chown,',
            'audit capability,',
            'capability dac_override,',
        ]

        expected_raw = [
            '  audit capability,',
            '',
        ]

        expected_clean = [
            '  audit capability,',
            '',
        ]

        expected_deleted = 2

        self._check_test_delete_duplicates_in_profile(rules, expected_raw, expected_clean, expected_deleted)

    def test_delete_duplicates_in_profile_02(self):
        rules = [
            'audit capability chown,',
            'audit capability,',
            'audit capability dac_override,',
            'capability ,',
            'audit capability ,',
        ]

        expected_raw = [
            '  audit capability,',
            '',
        ]

        expected_clean = [
            '  audit capability,',
            '',
        ]

        expected_deleted = 4

        self._check_test_delete_duplicates_in_profile(rules, expected_raw, expected_clean, expected_deleted)

    def test_delete_duplicates_in_profile_03(self):
        rules = [
            'audit capability chown,',
            'capability dac_override,',
            'deny capability dac_override,',
            'capability dac_override,',
            'audit capability chown,',
            'deny capability chown,',
            'audit deny capability chown,',
            'capability,',
            'audit capability,',
        ]

        expected_raw = [
            '  deny capability dac_override,',
            '  audit deny capability chown,',
            '  audit capability,',
            '',
        ]

        expected_clean = [
            '  audit deny capability chown,',
            '  deny capability dac_override,',
            '',
            '  audit capability,',
            '',
        ]

        expected_deleted = 6

        self._check_test_delete_duplicates_in_profile(rules, expected_raw, expected_clean, expected_deleted)

    def test_delete_duplicates_in_profile_04(self):
        rules = [
            'audit capability chown,',
            'deny capability chown,',
        ]

        expected_raw = [
            '  audit capability chown,',
            '  deny capability chown,',
            '',
        ]

        expected_clean = [
            '  deny capability chown,',
            '',
            '  audit capability chown,',
            '',
        ]

        expected_deleted = 0

        self._check_test_delete_duplicates_in_profile(rules, expected_raw, expected_clean, expected_deleted)


setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)



#!/usr/bin/python3
# ----------------------------------------------------------------------
#    Copyright (C) 2020 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import unittest
from collections import namedtuple
from common_test import AATest, setup_all_loops

from apparmor.common import AppArmorBug, AppArmorException
from apparmor.rule.abi import AbiRule, AbiRuleset
from apparmor.translations import init_translation

_ = init_translation()

exp = namedtuple(
    'exp', (  # 'audit', 'allow_keyword', 'deny',
        'comment', 'path', 'ifexists', 'ismagic'))

# --- tests for single AbiRule --- #


class AbiTest(AATest):
    def _compare_obj(self, obj, expected):
        self.assertEqual(False, obj.allow_keyword)  # not supported in abi rules, expected to be always False
        self.assertEqual(False, obj.audit)          # not supported in abi rules, expected to be always False
        self.assertEqual(False, obj.deny)           # not supported in abi rules, expected to be always False
        self.assertEqual(expected.comment, obj.comment)

        self.assertEqual(expected.path, obj.path)
        self.assertEqual(False, expected.ifexists)  # tests bug - should always expect ifexists==False
        self.assertEqual(False, obj.ifexists)       # not supported in abi rules, expected to be always False
        self.assertEqual(expected.ismagic, obj.ismagic)


class AbiTestParse(AbiTest):
    tests = (
        # AbiRule object                           comment       path                 if exists  ismagic
        ('abi <abstractions/base>,',           exp('',           'abstractions/base', False,     True)),  # magic path
        ('abi <abstractions/base>, # comment', exp(' # comment', 'abstractions/base', False,     True)),
        ('abi<abstractions/base>,#comment',    exp(' #comment',  'abstractions/base', False,     True)),
        ('   abi     <abstractions/base> , ',  exp('',           'abstractions/base', False,     True)),
        ('abi "/foo/bar",',                    exp('',           '/foo/bar',          False,     False)),  # absolute path
        ('abi "/foo/bar", # comment',          exp(' # comment', '/foo/bar',          False,     False)),
        ('abi "/foo/bar",#comment',            exp(' #comment',  '/foo/bar',          False,     False)),
        ('   abi "/foo/bar" , ',               exp('',           '/foo/bar',          False,     False)),
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(AbiRule.match(rawrule))
        obj = AbiRule.create_instance(rawrule)
        self.assertEqual(rawrule.strip(), obj.raw_rule)
        self._compare_obj(obj, expected)


class AbiTestParseInvalid(AbiTest):
    # tests = (
    #     (' some abi <abstractions/base>', AppArmorException),
    #     ('  /etc/fstab r,',               AppArmorException),
    #     ('/usr/abi r,',                   AppArmorException),
    #     ('/abi r,',                       AppArmorException),
    # )

    def _run_test(self, rawrule, expected):
        self.assertTrue(AbiRule.match(rawrule))  # the above invalid rules still match the main regex!
        with self.assertRaises(expected):
            AbiRule.create_instance(rawrule)

# class AbiTestParseFromLog(AbiTest):  # we'll never have log events for abi


class AbiFromInit(AbiTest):
    tests = (
        # AbiRule object     ifexists  ismagic                      comment  path        ifexists  ismagic
        (AbiRule('abi/4.19', False,    False),                  exp('',      'abi/4.19', False,    False)),
        (AbiRule('foo',      False,    False),                  exp('',      'foo',      False,    False)),
        (AbiRule('bar',      False,    True),                   exp('',      'bar',      False,    True)),
        (AbiRule('comment',  False,    False, comment='# cmt'), exp('# cmt', 'comment',  False,    False)),
    )

    def _run_test(self, obj, expected):
        self._compare_obj(obj, expected)


class InvalidAbiInit(AATest):
    tests = (
        # init params            expected exception
        ((False,  False, False), AppArmorBug),  # wrong type for path
        (('',     False, False), AppArmorBug),  # empty path
        ((None,   False, False), AppArmorBug),  # wrong type for path
        # (('    ', False, False), AppArmorBug),  # whitespace-only path
        (('foo',  None,  False), AppArmorBug),  # wrong type for ifexists
        (('foo',  '',    False), AppArmorBug),  # wrong type for ifexists
        (('foo',  False, None),  AppArmorBug),  # wrong type for ismagic
        (('foo',  False, ''),    AppArmorBug),  # wrong type for ismagic
        (('',     True,  False), AppArmorBug),  # ifexists set
    )

    def _run_test(self, params, expected):
        with self.assertRaises(expected):
            AbiRule(*params)

    def test_missing_params_1(self):
        with self.assertRaises(TypeError):
            AbiRule()

    def test_missing_params_2(self):
        with self.assertRaises(TypeError):
            AbiRule('foo')

    def test_missing_params_3(self):
        with self.assertRaises(TypeError):
            AbiRule('foo', False)

    def test_audit_true(self):
        with self.assertRaises(AppArmorBug):
            AbiRule('foo', False, False, audit=True)

    def test_deny_true(self):
        with self.assertRaises(AppArmorBug):
            AbiRule('foo', False, False, deny=True)

    def test_ifexists_true(self):
        with self.assertRaises(AppArmorBug):
            AbiRule('foo', True, False)


class InvalidAbiTest(AATest):
    def _check_invalid_rawrule(self, rawrule, matches_regex=False):
        obj = None
        self.assertEqual(AbiRule.match(rawrule), matches_regex)
        with self.assertRaises(AppArmorException):
            obj = AbiRule.create_instance(rawrule)

        self.assertIsNone(obj, 'AbiRule handed back an object unexpectedly')

    def test_invalid_abi_missing_path(self):
        self._check_invalid_rawrule('abi ,', matches_regex=True)  # missing path

    def test_invalid_non_AbiRule(self):
        self._check_invalid_rawrule('dbus,')  # not a abi rule

#   def test_empty_data_1(self):
#       obj = AbiRule('foo', False, False)
#       obj.path = ''
#       # no path set
#       with self.assertRaises(AppArmorBug):
#           obj.get_clean(1)


class WriteAbiTestAATest(AATest):
    def _run_test(self, rawrule, expected):
        self.assertTrue(AbiRule.match(rawrule))
        obj = AbiRule.create_instance(rawrule)
        clean = obj.get_clean()
        raw = obj.get_raw()

        self.assertEqual(expected.strip(), clean, 'unexpected clean rule')
        self.assertEqual(rawrule.strip(), raw, 'unexpected raw rule')

    tests = (
        #  raw rule                          clean rule
        ('     abi      <foo>   ,        ',  'abi <foo>,'),
        ('     abi       foo    ,        ',  'abi "foo",'),
        ('     abi      "foo"   ,        ',  'abi "foo",'),
        ('     abi       /foo   ,         ', 'abi "/foo",'),
        ('     abi      "/foo"  ,         ', 'abi "/foo",'),

        ('     abi      <foo>, # bar     ',  'abi <foo>, # bar'),
        ('     abi       foo , # bar     ',  'abi "foo", # bar'),
        ('     abi      "foo", # bar     ',  'abi "foo", # bar'),
        ('     abi       /foo,  # bar     ', 'abi "/foo", # bar'),
        ('     abi      "/foo", # bar     ', 'abi "/foo", # bar'),
    )

    def test_write_manually(self):
        obj = AbiRule('abs/foo', False, True, comment=' # cmt')

        expected = '    abi <abs/foo>, # cmt'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')


class AbiCoveredTest(AATest):
    def _run_test(self, param, expected):
        obj = AbiRule.create_instance(self.rule)
        check_obj = AbiRule.create_instance(param)

        self.assertTrue(AbiRule.match(param))

        self.assertEqual(obj.is_equal(check_obj), expected[0], 'Mismatch in is_equal, expected {}'.format(expected[0]))
        self.assertEqual(obj.is_equal(check_obj, True), expected[1], 'Mismatch in is_equal/strict, expected {}'.format(expected[1]))

        self.assertEqual(obj.is_covered(check_obj), expected[2], 'Mismatch in is_covered, expected {}'.format(expected[2]))
        self.assertEqual(obj.is_covered(check_obj, True, True), expected[3], 'Mismatch in is_covered/exact, expected {}'.format(expected[3]))


class AbiCoveredTest_01(AbiCoveredTest):
    rule = 'abi <foo>,'

    tests = (
        #   rule           equal  strict equal  covered  covered exact
        ('abi <foo>,',    (True,  True,         True,    True)),
        ('abi "foo",',    (False, False,        False,   False)),
        ('abi <foobar>,', (False, False,        False,   False)),
        ('abi "foo",',    (False, False,        False,   False)),
    )


class AbiCoveredTest_02(AbiCoveredTest):
    rule = 'abi "foo",'

    tests = (
        #   rule           equal  strict equal  covered  covered exact
        ('abi <foo>,',    (False, False,        False,   False)),
        ('abi "foo",',    (True,  True,         True,    True)),
        ('abi "foobar",', (False, False,        False,   False)),
        ('abi foo,',      (True,  False,        True,    True)),
    )


# class AbiCoveredTest_Invalid(AATest):
#     def test_borked_obj_is_covered_1(self):
#         obj = AbiRule.create_instance('abi <foo>')
#
#         testobj = AbiRule('foo', True, True)
#         testobj.path = ''
#
#         with self.assertRaises(AppArmorBug):
#             obj.is_covered(testobj)
#
#     def test_borked_obj_is_covered_2(self):
#         obj = AbiRule.create_instance('abi send set=quit peer=/foo,')
#
#         testobj = AbiRule('send', 'quit', '/foo')
#         testobj.abi = ''
#
#         with self.assertRaises(AppArmorBug):
#             obj.is_covered(testobj)
#
#     def test_borked_obj_is_covered_3(self):
#         obj = AbiRule.create_instance('abi send set=quit peer=/foo,')
#
#         testobj = AbiRule('send', 'quit', '/foo')
#         testobj.peer = ''
#
#         with self.assertRaises(AppArmorBug):
#             obj.is_covered(testobj)
#
#     def test_invalid_is_covered(self):
#         raw_rule = 'abi send,'
#         class SomeOtherClass(AbiRule):
#             pass
#
#         obj = AbiRule.create_instance(raw_rule)
#         testobj = SomeOtherClass.create_instance(raw_rule)  # different type
#         with self.assertRaises(AppArmorBug):
#             obj.is_covered(testobj)
#
#     def test_invalid_is_equal(self):
#         raw_rule = 'abi send,'
#         class SomeOtherClass(AbiRule):
#             pass
#
#         obj = AbiRule.create_instance(raw_rule)
#         testobj = SomeOtherClass.create_instance(raw_rule)  # different type
#         with self.assertRaises(AppArmorBug):
#             obj.is_equal(testobj)


class AbiLogprofHeaderTest(AATest):
    tests = (
        ('abi <abi/3.0>,',  [_('Abi'), 'abi <abi/3.0>,']),
        ('abi "/foo/bar",', [_('Abi'), 'abi "/foo/bar",']),
    )

    def _run_test(self, params, expected):
        obj = AbiRule.create_instance(params)
        self.assertEqual(obj.logprof_header(), expected)


# --- tests for AbiRuleset --- #

class AbiRulesTest(AATest):
    def test_empty_ruleset(self):
        ruleset = AbiRuleset()
        ruleset_2 = AbiRuleset()
        self.assertEqual([], ruleset.get_raw(2))
        self.assertEqual([], ruleset.get_clean(2))
        self.assertEqual([], ruleset_2.get_raw(2))
        self.assertEqual([], ruleset_2.get_clean(2))
        self.assertEqual([], ruleset_2.get_clean_unsorted(2))

    def test_ruleset_1(self):
        ruleset = AbiRuleset()
        rules = (
            ' abi  <foo>  ,',
            ' abi   "/bar", ',
        )

        expected_raw = [
            'abi  <foo>  ,',
            'abi   "/bar",',
            '',
        ]

        expected_clean = [
            'abi "/bar",',
            'abi <foo>,',
            '',
        ]

        expected_clean_unsorted = [
            'abi <foo>,',
            'abi "/bar",',
            '',
        ]

        for rule in rules:
            ruleset.add(AbiRule.create_instance(rule))

        self.assertEqual(expected_raw, ruleset.get_raw())
        self.assertEqual(expected_clean, ruleset.get_clean())
        self.assertEqual(expected_clean_unsorted, ruleset.get_clean_unsorted())


class AbiGlobTestAATest(AATest):
    def setUp(self):
        self.maxDiff = None
        self.ruleset = AbiRuleset()

    # def test_glob(self):
    #     with self.assertRaises(NotImplementedError):
    #         # get_glob_ext is not available for include rules
    #         self.ruleset.get_glob('include send set=int,')

    def test_glob_ext(self):
        with self.assertRaises(NotImplementedError):
            # get_glob_ext is not available for include rules
            self.ruleset.get_glob_ext('include send set=int,')


# class AbiDeleteTestAATest(AATest):
#     pass


setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)



# ----------------------------------------------------------------------
#    Copyright (C) 2013 Kshitij Gupta <kgupta8592@gmail.com>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------
import os
import shutil
import subprocess
import sys
import unittest

import apparmor.aa as apparmor
from common_test import AATest, read_file, write_file, setup_aa, setup_all_loops, skip_active_profiles

# Use the same python as the one this script is being run with
python_interpreter = sys.executable
if not python_interpreter:
    python_interpreter = 'python3'


class MinitoolsTest(AATest):

    def AASetup(self):
        self.createTmpdir()

        # copy the local profiles to the test directory
        # Should be the set of cleanprofile
        self.profile_dir = self.tmpdir + '/profiles'
        shutil.copytree('../../profiles/apparmor.d/', self.profile_dir,
                        symlinks=True, ignore=shutil.ignore_patterns(*skip_active_profiles))

        apparmor.profile_dir = self.profile_dir

        # Path for the program
        self.test_path = '/usr/sbin/winbindd'
        # Path for the target file containing profile
        self.local_profilename = self.profile_dir + '/usr.sbin.winbindd'

    def test_audit(self):
        # Set test profile to audit mode and check if it was correctly set
        subprocess.check_output(
            '{} ./../aa-audit --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)

        self.assertEqual(
            apparmor.get_profile_flags(self.local_profilename, self.test_path),
            'audit',
            'Audit flag could not be set in profile ' + self.local_profilename)

        # Remove audit mode from test profile and check if it was correctly removed
        subprocess.check_output(
            '{} ./../aa-audit --no-reload -d {} -r {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path), shell=True)

        self.assertEqual(
            apparmor.get_profile_flags(self.local_profilename, self.test_path),
            None,
            'Audit flag could not be removed in profile ' + self.local_profilename)

    def test_complain(self):
        # Set test profile to complain mode and check if it was correctly set
        subprocess.check_output(
            '{} ./../aa-complain --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)

        # "manually" create a force-complain symlink (will be deleted by aa-enforce later)
        force_complain_dir = self.profile_dir + '/force-complain'
        if not os.path.isdir(force_complain_dir):
            os.mkdir(force_complain_dir)
        os.symlink(
            self.local_profilename,
            '{}/{}'.format(force_complain_dir, os.path.basename(self.local_profilename)))

        self.assertEqual(
            os.path.islink('{}/{}'.format(force_complain_dir, os.path.basename(self.local_profilename))),
            True,
            'Failed to create a symlink for {} in force-complain'.format(self.local_profilename))
        self.assertEqual(
            apparmor.get_profile_flags(self.local_profilename, self.test_path),
            'complain',
            'Complain flag could not be set in profile ' + self.local_profilename)

        # Set test profile to enforce mode and check if it was correctly set
        subprocess.check_output(
            '{} ./../aa-enforce --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)

        self.assertEqual(
            os.path.islink('{}/{}'.format(force_complain_dir, os.path.basename(self.local_profilename))),
            False,
            'Failed to remove symlink for {} from force-complain'.format(self.local_profilename))
        self.assertEqual(
            os.path.islink('{}/disable/{}'.format(self.profile_dir, os.path.basename(self.local_profilename))),
            False,
            'Failed to remove symlink for {} from disable'.format(self.local_profilename))
        self.assertEqual(
            apparmor.get_profile_flags(self.local_profilename, self.test_path),
            None,
            'Complain flag could not be removed in profile ' + self.local_profilename)

        # Set audit flag and then complain flag in a profile
        subprocess.check_output(
            '{} ./../aa-audit --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)
        subprocess.check_output(
            '{} ./../aa-complain --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)
        # "manually" create a force-complain symlink (will be deleted by aa-enforce later)
        os.symlink(
            self.local_profilename,
            '{}/{}'.format(force_complain_dir, os.path.basename(self.local_profilename)))

        self.assertEqual(
            os.path.islink('{}/{}'.format(force_complain_dir, os.path.basename(self.local_profilename))),
            True,
            'Failed to create a symlink for {} in force-complain'.format(self.local_profilename))
        self.assertEqual(
            apparmor.get_profile_flags(self.local_profilename, self.test_path),
            'audit, complain',
            'Complain flag could not be set in profile ' + self.local_profilename)

        # Remove complain flag first i.e. set to enforce mode
        subprocess.check_output(
            '{} ./../aa-enforce --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)

        self.assertEqual(
            os.path.islink('{}/{}'.format(force_complain_dir, os.path.basename(self.local_profilename))),
            False,
            'Failed to remove symlink for {} from force-complain'.format(self.local_profilename))
        self.assertEqual(
            os.path.islink('{}/disable/{}'.format(self.profile_dir, os.path.basename(self.local_profilename))),
            False,
            'Failed to remove symlink for {} from disable'.format(self.local_profilename))
        self.assertEqual(
            apparmor.get_profile_flags(self.local_profilename, self.test_path),
            'audit',
            'Complain flag could not be removed in profile ' + self.local_profilename)

        # Remove audit flag
        subprocess.check_output(
            '{} ./../aa-audit --no-reload -d {} -r {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)

    def test_enforce(self):
        # Set test profile to enforce mode and check if it was correctly set
        subprocess.check_output(
            '{} ./../aa-enforce --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)

        self.assertEqual(
            os.path.islink('{}/force-complain/{}'.format(self.profile_dir, os.path.basename(self.local_profilename))),
            False,
            'Failed to remove symlink for {} from force-complain'.format(self.local_profilename))
        self.assertEqual(
            os.path.islink('{}/disable/{}'.format(self.profile_dir, os.path.basename(self.local_profilename))),
            False,
            'Failed to remove symlink for {} from disable'.format(self.local_profilename))
        self.assertEqual(
            apparmor.get_profile_flags(self.local_profilename, self.test_path),
            None,
            'Complain flag could not be removed in profile {}'.format(self.local_profilename))

    def test_disable(self):
        # Disable the test profile and check if it was correctly disabled
        subprocess.check_output(
            '{} ./../aa-disable --no-reload -d {} {} --configdir ./'.format(
                python_interpreter, self.profile_dir, self.test_path),
            shell=True)

        self.assertEqual(
            os.path.islink('{}/disable/{}'.format(self.profile_dir, os.path.basename(self.local_profilename))),
            True,
            'Failed to create a symlink for {} in disable'.format(self.local_profilename))

    def test_autodep(self):
        # small bash script - we'll create a profile for it with aa-autodep
        script_filename = write_file(self.tmpdir, 'autodep_test.sh', '#!/bin/bash\necho hello world')

        subprocess.check_output([python_interpreter, './../aa-autodep', '--no-reload', '-d', self.profile_dir, '--configdir',  './', script_filename])

        expected_file = apparmor.get_new_profile_filename(script_filename)

        self.assertTrue(os.path.isfile(expected_file), 'Failed to create profile with aa-autodep: {}'.format(script_filename))

        prof_content = read_file(expected_file).split('\n')

        self.assertTrue(prof_content[0].startswith('# Last Modified:'), 'prof_content[0] starts with %s' % prof_content[0])
        self.assertEqual(prof_content[1], 'abi <abi/4.0>,')
        self.assertEqual(prof_content[2], '')
        self.assertEqual(prof_content[3], 'include <tunables/global>')
        self.assertEqual(prof_content[4], '')
        self.assertEqual(prof_content[5], '%s flags=(complain) {' % script_filename)
        self.assertEqual(prof_content[6], '  include <abstractions/base>')
        self.assertEqual(prof_content[7], '  include <abstractions/bash>')
        self.assertEqual(prof_content[8], '')
        self.assertEqual(prof_content[9], '  %s r,' % script_filename)
        self.assertTrue(prof_content[10] in ['  /usr/bin/bash ix,', '  /bin/bash ix,'])
        self.assertEqual(prof_content[11], '')
        self.assertEqual(prof_content[12], '}')
        self.assertEqual(prof_content[13], '')

        with self.assertRaises(IndexError):
            prof_content[14]

    @unittest.skipIf(apparmor.check_for_apparmor() is None, "Securityfs not mounted or doesn't have the apparmor directory.")
    def test_unconfined(self):
        output = subprocess.check_output(
            python_interpreter + ' ./../aa-unconfined --configdir ./', shell=True)

        output_force = subprocess.check_output(
            python_interpreter + ' ./../aa-unconfined --paranoid --configdir ./', shell=True)

        self.assertIsNot(output, '', 'Failed to run aa-unconfined')

        self.assertIsNot(output_force, '', 'Failed to run aa-unconfined in paranoid mode')

    def test_cleanprof(self):
        input_file = 'cleanprof_test.in'
        output_file = 'cleanprof_test.out'
        # We position the local testfile
        shutil.copy('./' + input_file, self.profile_dir)
        # Our silly test program whose profile we wish to clean
        cleanprof_test = '/usr/bin/a/simple/cleanprof/test/profile'

        subprocess.check_output(
            '{} ./../aa-cleanprof  --no-reload -d {} -s {} --configdir ./'.format(
                python_interpreter, self.profile_dir, cleanprof_test),
            shell=True)

        # Strip off the first line (#modified line)
        subprocess.check_output('sed -i 1d {}/{}'.format(self.profile_dir, input_file), shell=True)

        exp_content = read_file('./' + output_file)
        real_content = read_file('{}/{}'.format(self.profile_dir, input_file))
        self.maxDiff = None
        self.assertEqual(exp_content, real_content, 'Failed to cleanup profile properly')


setup_aa(apparmor)
setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)



#! /usr/bin/python3
# ------------------------------------------------------------------
#
#    Copyright (C) 2015-2021 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License published by the Free Software Foundation.
#
# ------------------------------------------------------------------

import os
import sys
import unittest

import apparmor.aa
from apparmor.common import hasher, open_file_read, split_name
from apparmor.logparser import ReadLog
from apparmor.profile_list import ProfileList
from common_test import AATest, read_file, setup_aa, setup_all_loops


class TestLibapparmorTestMulti(AATest):
    """Parse all libraries/libapparmor/testsuite/test_multi tests and compare the result with the *.out files"""

    tests = 'invalid'  # filled by parse_test_profiles()

    def _run_test(self, params, expected):
        # tests[][expected] is a dummy, replace it with the real values
        if params.split('/')[-1] in log_to_skip:
            return

        expected = self._parse_libapparmor_test_multi(params)

        loglines = []
        with open_file_read(params + '.in') as f_in:
            for line in f_in:
                if line.strip():
                    loglines.append(line)

        self.assertEqual(len(loglines), 1, params + '.in should only contain one line!')

        parser = ReadLog('', '', '')
        parsed_event = parser.parse_event(loglines[0])

        if parsed_event and expected:
            parsed_items = dict(parsed_event.items())

            # check if the line passes the regex in logparser.py
            if not parser.RE_LOG_ALL.search(loglines[0]):
                raise Exception("Log event doesn't match RE_LOG_ALL")

            for label in expected:
                if label in (
                        'file',  # filename of the *.in file
                        'event_type',  # mapped to aamode
                        'audit_id', 'audit_sub_id',  # not set nor relevant
                        'comm',  # not set, and not too useful
                        # XXX most of the keywords listed below mean "TODO"
                        'fsuid', 'ouid',  # file events
                        'flags', 'fs_type',  # mount
                        'namespace',  # file_lock only?? (at least the tests don't contain this in other event types with namespace)
                        'net_local_addr', 'net_foreign_addr', 'net_local_port', 'net_foreign_port',  # detailed network events
                        'peer', 'signal',  # signal
                        'src_name',  # pivotroot
                        'dbus_bus', 'dbus_interface', 'dbus_member', 'dbus_path',  # dbus
                        'peer_pid', 'peer_profile',  # dbus
                        'net_addr', 'peer_addr',  # unix

                ):
                    pass
                elif parsed_items['operation'] == 'exec' and label in ('sock_type', 'family', 'protocol'):
                    pass  # XXX 'exec' + network? really?
                elif parsed_items['operation'] == 'ptrace' and label == 'name2' and params.endswith('/ptrace_garbage_lp1689667_1'):
                    pass  # libapparmor would better qualify this case as invalid event
                elif not parsed_items.get(label, None):
                    raise Exception('parsed_items[{}] not set'.format(label))
                elif not expected.get(label, None):
                    raise Exception('expected[{}] not set'.format(label))
                else:
                    self.assertEqual(str(parsed_items[label]), expected[label], label + ' differs')
        elif expected:
            self.assertIsNone(parsed_event)  # that's why we end up here
            self.assertEqual(dict(), expected, 'parsed_event is none')  # effectively print the content of expected
        elif parsed_event:
            self.assertIsNone(expected)  # that's why we end up here
            self.assertEqual(parsed_event, dict(), 'expected is none')  # effectively print the content of parsed_event
        else:
            self.assertIsNone(expected)  # that's why we end up here
            self.assertIsNone(parsed_event)  # that's why we end up here
            self.assertEqual(parsed_event, expected)  # both are None

    # list of labels that use a different name in logparser.py than in the test_multi *.out files
    # (additionally, .lower() is applied to all labels)
    label_map = {
        'Mask': 'request_mask',
        'Command': 'comm',
        'Token': 'magic_token',
        'ErrorCode': 'error_code',
        'Network family': 'family',
        'Socket type': 'sock_type',
        'Local addr': 'net_local_addr',
        'Foreign addr': 'net_foreign_addr',
        'Local port': 'net_local_port',
        'Foreign port': 'net_foreign_port',
        'Audit subid': 'audit_sub_id',
        'Attribute': 'attr',
        'Epoch': 'time',
    }

    def _parse_libapparmor_test_multi(self, file_with_path):
        """parse the libapparmor test_multi *.in tests and their expected result in *.out"""

        with open_file_read(file_with_path + '.out') as f_in:
            expected = f_in.readlines()

        if expected[0].rstrip('\n') != 'START':
            raise Exception("{}.out doesn't have 'START' in its first line! ({})".format(
                file_with_path, expected[0]))

        expected.pop(0)

        exresult = dict()
        for line in expected:
            label, value = line.split(':', 1)

            # test_multi doesn't always use the original labels :-/
            if label in self.label_map.keys():
                label = self.label_map[label]
            label = label.replace(' ', '_').lower()
            exresult[label] = value.strip()

        if not exresult['event_type'].startswith('AA_RECORD_'):
            raise Exception("event_type doesn't start with AA_RECORD_: {} in file {}".format(
                exresult['event_type'], file_with_path))

        exresult['aamode'] = exresult['event_type'].replace('AA_RECORD_', '')
        if exresult['aamode'] == 'ALLOWED':
            exresult['aamode'] = 'PERMITTING'
        if exresult['aamode'] == 'DENIED':
            exresult['aamode'] = 'REJECTING'

        if exresult['event_type'] == 'AA_RECORD_INVALID':  # or exresult.get('error_code', 0) != 0:  # XXX should events with errors be ignored?
            exresult = None

        return exresult


# tests that cause crashes or need user interaction (will be skipped)
log_to_skip = [
    'testcase_dbus_09',  # multiline log not currently supported
]

# tests that do not produce the expected profile (checked with assertNotEqual)
log_to_profile_known_failures = [
    # exec events
    'testcase01',
    'testcase12',
    'testcase13',
]

# tests that cause crashes or need user interaction (will be skipped)
log_to_profile_skip = [
    'testcase31',  # XXX AppArmorBug: Log contains unknown mode mrwIxl

    'testcase_dmesg_changehat_negative_error',   # fails in write_header -> quote_if_needed because data is None
    'testcase_syslog_changehat_negative_error',  # fails in write_header -> quote_if_needed because data is None

    'testcase_changehat_01',  # interactive, asks to add a hat
    'testcase_dbus_09',  # multiline log not currently supported
]

# tests that cause an empty log
log_to_profile_known_empty_log = [
    'change_onexec_lp1648143',  # change_onexec not supported in logparser.py yet (and the log is about "no new privs" error)
    'ptrace_garbage_lp1689667_1',  # no denied= in log
    'ptrace_no_denied_mask',  # no denied= in log
    'unconfined-change_hat',  # unconfined trying to change_hat, which isn't allowed
    'testcase_unix_no_denied_issue541',  # no denied_mask in log
]


class TestLogToProfile(AATest):
    """Check if the libraries/libapparmor/testsuite/test_multi tests result in the expected profile"""

    tests = 'invalid'  # filled by parse_test_profiles()

    def _run_test(self, params, expected):
        logfile = params + '.in'

        if params.split('/')[-1] in log_to_profile_skip:
            return

        profile, new_profile = logfile_to_profile(logfile)
        if profile is None:
            return

        expected_profile = read_file(params + '.profile')

        if params.split('/')[-1] in log_to_profile_known_failures:
            self.assertNotEqual(new_profile, expected_profile)  # known failure
        else:
            self.assertEqual(new_profile, expected_profile)


def logfile_to_profile(logfile):
    profile_dummy_file = 'AATest_does_exist'

    # we need to find out the profile name and aamode (complain vs. enforce mode) so that the test can access the correct place in storage
    parser = ReadLog('', '', '')
    parsed_event = parser.parse_event(read_file(logfile))

    if not parsed_event:  # AA_RECORD_INVALID
        return None, 'INVALID'

    aamode = parsed_event['aamode']

    if aamode in ('AUDIT', 'STATUS', 'HINT'):  # ignore some event types  # XXX maybe we shouldn't ignore AUDIT events?
        return None, aamode

    if aamode not in ('PERMITTING', 'REJECTING'):
        raise Exception('Unexpected aamode {}'.format(parsed_event['aamode']))

    # cleanup apparmor.aa storage
    apparmor.aa.log = dict()
    apparmor.aa.aa = hasher()

    profile, hat = split_name(parsed_event['profile'])

    apparmor.aa.active_profiles = ProfileList()

    dummy_prof = apparmor.aa.ProfileStorage('TEST DUMMY for active_profiles', profile_dummy_file, 'logprof_to_profile()')

    # optional for now, might be needed one day
    # if profile.startswith('/'):
    #     apparmor.aa.active_profiles.add_profile(profile_dummy_file, profile, profile, dummy_prof)
    # else:
    apparmor.aa.active_profiles.add_profile(profile_dummy_file, profile, '', dummy_prof)

    log_reader = ReadLog(logfile, apparmor.aa.active_profiles, '')
    hashlog = log_reader.read_log('')

    apparmor.aa.ask_exec(hashlog)
    apparmor.aa.ask_addhat(hashlog)

    log_dict = apparmor.aa.collapse_log(hashlog, ignore_null_profiles=False)

    if list(log_dict[aamode].keys()) != [parsed_event['profile']]:
        raise Exception('log_dict[{}] contains unexpected keys. Logfile: {}, keys {}'.format(aamode, logfile, log_dict.keys()))

    if '//' in parsed_event['profile']:
        # log event for a child profile means log_dict only contains the child profile
        # initialize parent profile in log_dict as ProfileStorage to ensure writing the profile doesn't fail
        # (in "normal" usage outside of this test, log_dict will not be handed over to serialize_profile())

        log_dict[aamode][profile] = apparmor.aa.ProfileStorage('TEST DUMMY for empty parent profile', profile_dummy_file, 'logfile_to_profile()')
        log_dict[aamode][parsed_event['profile']]['is_hat'] = True  # for historical reasons, generate hats, not child profiles

    log_is_empty = True

    for tmpaamode in hashlog:
        for tmpprofile in hashlog[tmpaamode]:
            for tmpruletype in hashlog[tmpaamode][tmpprofile]:
                if tmpruletype == 'final_name' and hashlog[tmpaamode][tmpprofile]['final_name'] == tmpprofile:
                    continue  # final_name is a copy of the profile name (may be changed by ask_exec(), but that won't happen in this test)
                if hashlog[tmpaamode][tmpprofile][tmpruletype]:
                    log_is_empty = False

    if logfile.split('/')[-1][:-3] in log_to_profile_known_empty_log:
        # unfortunately this function might be called outside Unittest.TestCase, therefore we can't use assertEqual / assertNotEqual
        if not log_is_empty:
            raise Exception('got non-empty log for logfile in log_to_profile_known_empty_log: {} {}'.format(logfile, hashlog))
    else:
        if log_is_empty:
            raise Exception('got empty log for logfile not in log_to_profile_known_empty_log: {} {}'.format(logfile, hashlog))

    new_profile = apparmor.aa.serialize_profile(log_dict[aamode], profile, {})

    return profile, new_profile


def find_test_multi(log_dir):
    """find all log sniplets in the given log_dir"""

    log_dir = os.path.abspath(log_dir)

    tests = []
    for root, dirs, files in os.walk(log_dir):
        for file in files:
            if file.endswith('.in'):
                file_with_path = os.path.join(root, file[:-3])  # filename without '.in'
                tests.append((file_with_path, True))  # True is a dummy testresult, parsing of the *.out files is done while running the tests

            elif file.endswith('.out') or file.endswith('.err') or file.endswith('.profile'):
                pass
            else:
                raise Exception('Found unknown file {} in libapparmor test_multi'.format(file))

    return tests


# if a logfile is given as parameter, print the resulting profile and exit (with $? = 42 to make sure tests break if the caller accidentally hands over a parameter)
if __name__ == '__main__' and len(sys.argv) == 2:
    print(logfile_to_profile(sys.argv[1])[1])
    sys.exit(42)

# still here? That means a normal test run
print('Testing libapparmor test_multi tests...')
TestLibapparmorTestMulti.tests = find_test_multi('../../libraries/libapparmor/testsuite/test_multi/')
TestLogToProfile.tests = find_test_multi('../../libraries/libapparmor/testsuite/test_multi/')

setup_aa(apparmor.aa)
setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)



#!/usr/bin/python3
# ----------------------------------------------------------------------
#    Copyright (C) 2015 Christian Boltz <apparmor@cboltz.de>
#
#    This program is free software; you can redistribute it and/or
#    modify it under the terms of version 2 of the GNU General Public
#    License as published by the Free Software Foundation.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
# ----------------------------------------------------------------------

import unittest
from collections import namedtuple

import apparmor.severity as severity
from apparmor.common import AppArmorBug, AppArmorException, hasher
from apparmor.logparser import ReadLog
from apparmor.rule.file import FileRule, FileRuleset
from apparmor.translations import init_translation
from common_test import AATest, setup_all_loops

_ = init_translation()

exp = namedtuple(
    'exp', ('audit', 'allow_keyword', 'deny', 'comment', 'path', 'all_paths',
            'perms', 'all_perms', 'exec_perms', 'target', 'all_targets',
            'owner', 'file_keyword', 'leading_perms'))


# --- tests for single FileRule --- #

class FileTest(AATest):
    def _compare_obj(self, obj, expected):
        self.assertEqual(obj.allow_keyword, expected.allow_keyword)
        self.assertEqual(obj.audit, expected.audit)
        self.assertEqual(obj.deny, expected.deny)
        self.assertEqual(obj.comment, expected.comment)

        self._assertEqual_aare(obj.path, expected.path)
        self.assertEqual(obj.perms, expected.perms)
        self.assertEqual(obj.exec_perms, expected.exec_perms)
        self._assertEqual_aare(obj.target, expected.target)
        self.assertEqual(obj.owner, expected.owner)
        self.assertEqual(obj.file_keyword, expected.file_keyword)
        self.assertEqual(obj.leading_perms, expected.leading_perms)

        # Note: there's no all_ field for exec_perms, owner, file_keyword and leading_perms
        self.assertEqual(obj.all_paths, expected.all_paths)
        self.assertEqual(obj.all_perms, expected.all_perms)
        self.assertEqual(obj.all_targets, expected.all_targets)

    def _assertEqual_aare(self, obj, expected):
        if obj:
            self.assertEqual(obj.regex, expected)
        else:
            self.assertEqual(obj, expected)


class FileTestParse(FileTest):
    tests = (
        # FileRule object                    audit  allow  deny   comment   path               all_paths  perms               all?   exec_perms  target  all?   owner  file keyword  leading perms

        # bare file rules
        ('file,',                        exp(False, False, False, '',       None,              True,      None,               True,  None,       None,   True,  False, False,        False)),
        ('allow file,',                  exp(False, True,  False, '',       None,              True,      None,               True,  None,       None,   True,  False, False,        False)),
        ('audit deny owner file, # cmt', exp(True,  False, True,  ' # cmt', None,              True,      None,               True,  None,       None,   True,  True,  False,        False)),

        # "normal" file rules
        ('/foo r,',                      exp(False, False, False, '',       '/foo',            False,     {'r'},              False, None,       None,   True,  False, False,        False)),
        ('file /foo rwix,',              exp(False, False, False, '',       '/foo',            False,     {'r', 'w'},         False, 'ix',       None,   True,  False, True,         False)),
        ('/foo Px -> bar,',              exp(False, False, False, '',       '/foo',            False,     set(),              False, 'Px',       'bar',  False, False, False,        False)),
        ('@{PROC}/[a-z]** mr,',          exp(False, False, False, '',       '@{PROC}/[a-z]**', False,     {'r', 'm'},         False, None,       None,   True,  False, False,        False)),

        ('audit /tmp/foo r,',            exp(True,  False, False, '',       '/tmp/foo',        False,     {'r'},              False, None,       None,   True,  False, False,        False)),
        ('audit deny /tmp/foo r,',       exp(True,  False, True,  '',       '/tmp/foo',        False,     {'r'},              False, None,       None,   True,  False, False,        False)),
        ('audit deny /tmp/foo rx,',      exp(True,  False, True,  '',       '/tmp/foo',        False,     {'r'},              False, 'x',        None,   True,  False, False,        False)),
        ('allow /tmp/foo ra,',           exp(False, True,  False, '',       '/tmp/foo',        False,     {'r', 'a'},         False, None,       None,   True,  False, False,        False)),
        ('audit allow /tmp/foo ra,',     exp(True,  True,  False, '',       '/tmp/foo',        False,     {'r', 'a'},         False, None,       None,   True,  False, False,        False)),


        # file rules with leading permission
        ('r /foo,',                      exp(False, False, False, '',       '/foo',            False,     {'r'},              False, None,       None,   True,  False, False,        True)),
        ('file rwix /foo,',              exp(False, False, False, '',       '/foo',            False,     {'r', 'w'},         False, 'ix',       None,   True,  False, True,         True)),
        ('Px /foo -> bar,',              exp(False, False, False, '',       '/foo',            False,     set(),              False, 'Px',       'bar',  False, False, False,        True)),
        ('mr @{PROC}/[a-z]**,',          exp(False, False, False, '',       '@{PROC}/[a-z]**', False,     {'r', 'm'},         False, None,       None,   True,  False, False,        True)),

        ('audit r /tmp/foo,',            exp(True,  False, False, '',       '/tmp/foo',        False,     {'r'},              False, None,       None,   True,  False, False,        True)),
        ('audit deny r /tmp/foo,',       exp(True,  False, True,  '',       '/tmp/foo',        False,     {'r'},              False, None,       None,   True,  False, False,        True)),
        ('allow ra /tmp/foo,',           exp(False, True,  False, '',       '/tmp/foo',        False,     {'r', 'a'},         False, None,       None,   True,  False, False,        True)),
        ('audit allow ra /tmp/foo,',     exp(True,  True,  False, '',       '/tmp/foo',        False,     {'r', 'a'},         False, None,       None,   True,  False, False,        True)),

        # duplicated (but not conflicting) permissions
        ('/foo PxPxPxPxrwPx -> bar,',    exp(False, False, False, '',       '/foo',            False,     {'r', 'w'},         False, 'Px',       'bar',  False, False, False,        False)),
        ('/foo CixCixrwCix -> bar, ',    exp(False, False, False, '',       '/foo',            False,     {'r', 'w'},         False, 'Cix',      'bar',  False, False, False,        False)),

        # link rules
        ('link /foo -> /bar,',           exp(False, False, False, '',       '/foo',            False,     {'link'},           False, None,       '/bar', False, False, False,        True)),
        ('link subset /foo -> /bar,',    exp(False, False, False, '',       '/foo',            False,     {'link', 'subset'}, False, None,       '/bar', False, False, False,        True)),
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(FileRule.match(rawrule))
        obj = FileRule.create_instance(rawrule)
        self.assertEqual(rawrule.strip(), obj.raw_rule)
        self._compare_obj(obj, expected)


class FileTestParseInvalid(FileTest):
    tests = (
        ('/foo x,',       AppArmorException),  # should be *x
        ('/foo raw,',     AppArmorException),  # r and a conflict
        ('deny /foo ix,', AppArmorException),  # endy only allows x, but not *x
        ('deny /foo Px,', AppArmorException),  # deny only allows x, but not *x
        ('deny /foo Pi,', AppArmorException),  # missing 'x', and P not allowed
        ('allow /foo x,', AppArmorException),  # should be *x
        ('/foo Pxrix,',   AppArmorException),  # exec mode conflict
        ('/foo PixUx,',   AppArmorException),  # exec mode conflict
        ('/foo PxUx,',    AppArmorException),  # exec mode conflict
        ('/foo PUxPix,',  AppArmorException),  # exec mode conflict
        ('/foo Pi,',      AppArmorException),  # missing 'x'
    )

    def _run_test(self, rawrule, expected):
        self.assertTrue(FileRule.match(rawrule))  # the above invalid rules still match the main regex!
        with self.assertRaises(expected):
            FileRule.create_instance(rawrule)


class FileTestNonMatch(AATest):
    tests = (
        ('file /foo,',        False),
        ('file rw,',          False),
        ('file -> bar,',      False),
        ('file Px -> bar,',   False),
        ('/foo bar,',         False),
        ('dbus /foo,',        False),
        ('link /foo,',        False),  # missing '-> /target'
        ('link -> /bar,',     False),  # missing path
        ('/foo -> bar link,', False),  # link has to be leading keyword
        ('link,',             False),  # link isn't available as bare keyword
    )

    def _run_test(self, rawrule, expected):
        self.assertFalse(FileRule.match(rawrule))


class FileTestParseFromLog(FileTest):
    def test_file_from_log(self):
        parser = ReadLog('', '', '')
        event = 'Nov 11 07:33:07 myhost kernel: [50812.879558] type=1502 audit(1236774787.169:369): operation="inode_permission" requested_mask="::r" denied_mask="::r" fsuid=1000 name="/bin/dash" pid=13726 profile="/bin/foobar"'

        parsed_event = parser.parse_event(event)

        self.assertEqual(parsed_event, {
            'request_mask': '::r',
            'denied_mask': '::r',
            'error_code': 0,
            'magic_token': 0,
            'parent': 0,
            'profile': '/bin/foobar',
            'operation': 'inode_permission',
            'name': '/bin/dash',
            'name2': None,
            'resource': None,
            'info': None,
            'aamode': 'PERMITTING',
            'time': 1236774787,
            'active_hat': None,
            'pid': 13726,
            'task': 0,
            'attr': None,
            'family': None,
            'protocol': None,
            'sock_type': None,
            'class': None,
        })

        #     FileRule path,                 perms,                         exec_perms, target,       owner, file_keyword, leading_perms
        # obj = FileRule(parsed_event['name'], parsed_event['denied_mask'],   None,       FileRule.ALL, False, False,        False,)
        obj = FileRule(parsed_event['name'], 'r',                           None,       FileRule.ALL, False, False,        False,)
        # XXX handle things like '::r'
        # XXX split off exec perms?

        #              audit  allow  deny   comment  path          all_paths  perms  all?   exec_perms  target  all?   owner  file keyword  leading perms
        expected = exp(False, False, False, '',      '/bin/dash',  False,     {'r'}, False, None,       None,   True,  False, False,        False)

        self._compare_obj(obj, expected)

        self.assertEqual(obj.get_raw(1), '  /bin/dash r,')


# TODO: add logparser example for link event

class FileFromInit(FileTest):
    tests = (

        #         path,    perms, exec_perms, target,       owner, file_keyword, leading_perms
        (FileRule('/foo', 'rw',   None,       FileRule.ALL, False, False,        False,       audit=True, deny=True),
         #   audit  allow   deny  comment  path   all_paths perms       all?   exec_perms target all?  owner  file keyword  leading perms
         exp(True,  False, True, '',     '/foo', False,    {'r', 'w'}, False, None,      None,  True, False, False,        False)),

        #         path,   perms, exec_perms, target,     owner, file_keyword, leading_perms
        (FileRule('/foo', None,  'Pix',      'bar_prof', True,  True,         True,        allow_keyword=True),
         #   audit  allow  deny   comment  path     all_paths  perms  all?   exec_perms  target      all?   owner  file keyword  leading perms
         exp(False, True,  False, '',      '/foo',  False,     set(), False, 'Pix',      'bar_prof', False, True,  True,         True)),

        #         path,   perms,              exec_perms, target, owner, file_keyword, leading_perms
        (FileRule('/foo', {'link', 'subset'}, None,       '/bar', False, False,        True,        audit=True, deny=True),
         #   audit  allow  deny  comment  path    all_paths  perms               all?   exec_perms  target  all?   owner  file keyword leading perms
         exp(True,  False, True, '',      '/foo', False,     {'link', 'subset'}, False, None,       '/bar', False, False, False,       True)),

    )

    def _run_test(self, obj, expected):
        self._compare_obj(obj, expected)


class InvalidFileInit(AATest):
    tests = (
        #       path,         perms,              exec_perms, target, owner,  file_keyword, leading_perms  expected exception

        # empty fields
        (      ('',           'rw',               'ix',       '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        #  OK  ('/foo',       '',                 'ix',       '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               '',         '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '',     False,  False,        False),        AppArmorBug),  # noqa: E201

        # whitespace fields
        (      ('   ',        'rw',               'ix',       '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       '   ',              'ix',       '/bar', False,  False,        False),        AppArmorException),  # noqa: E201
        (      ('/foo',       'rw',               '   ',      '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '   ',  False,  False,        False),        AppArmorBug),  # noqa: E201

        # wrong type - dict()
        (      (dict(),       'rw',               'ix',       '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       dict(),             'ix',       '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               dict(),     '/bar', False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       dict(), False,  False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '/bar', dict(), False,        False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '/bar', False,  dict(),       False),        AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '/bar', False,  False,        dict()),       AppArmorBug),  # noqa: E201


        # wrong type - None
        (      (None,         'rw',               'ix',       '/bar', False,  False,        False),         AppArmorBug),  # noqa: E201
        #  OK  ('/foo',       None,               'ix',       '/bar', False,  False,        False),         AppArmorBug),  # noqa: E201
        #  OK  ('/foo',       'rw',               None,       '/bar', False,  False,        False),         AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       None,   False,  False,        False),         AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '/bar', None,   False,        False),         AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '/bar', False,  None,         False),         AppArmorBug),  # noqa: E201
        (      ('/foo',       'rw',               'ix',       '/bar', False,  False,        None),          AppArmorBug),  # noqa: E201


        # misc
        (      ('/foo',       'rwa',              'ix',       '/bar', False,  False,        False),        AppArmorException),  # 'r' and 'a' conflict  # noqa: E201
        (      ('/foo',       None,               'rw',       '/bar', False,  False,        False),        AppArmorBug),        # file perms in exec perms parameter  # noqa: E201
        (      ('/foo',       'ix',               None,       '/bar', False,  False,        False),        AppArmorBug),        # exec perms in file perms parameter  # noqa: E201
        (      ('foo',        'rw',               'ix',       '/bar', False,  False,        False),        AppArmorException),  # path doesn't start with /  # noqa: E201
        (      ('/foo',       'rb',               'ix',       '/bar', False,  False,        False),        AppArmorException),  # invalid file mode 'b' (str)  # noqa: E201
        (      ('/foo',       {'b'},              'ix',       '/bar', False,  False,        False),        AppArmorBug),        # invalid file mode 'b' (str)  # noqa: E201
        (      ('/foo',       'rw',               'ax',       '/bar', False,  False,        False),        AppArmorBug),        # invalid exec mode 'ax'  # noqa: E201
        (      ('/foo',       'rw',               'x',        '/bar', False,  False,        False),        AppArmorException),  # plain 'x' is only allowed in deny rules  # noqa: E201
        (      (FileRule.ALL, FileRule.ALL,       None,       '/bar', False,  False,        False),        AppArmorBug),        # plain 'file,' doesn't allow exec target  # noqa: E201

        # link rules
        (      (None,         {'link'},           None,       None,   False,  False,        False),        AppArmorBug),        # missing path and target  # noqa: E201
        (      ('/foo',       {'link'},           None,       None,   False,  False,        False),        AppArmorBug),        # missing target  # noqa: E201
        (      ( None,        {'link'},           None,       '/bar', False,  False,        False),        AppArmorBug),        # missing path  # noqa: E201
        (      ('/foo',       {'subset'},         None,       '/bar', False,  False,        False),        AppArmorBug),        # subset without link  # noqa: E201
        (      ('/foo',       {'link'},           'ix',       '/bar', False,  False,        False),        AppArmorBug),        # link rule with exec perms  # noqa: E201
        (      ('/foo',       {'link', 'subset'}, 'ix',       '/bar', False,  False,        False),        AppArmorBug),        # link subset rule with exec perms  # noqa: E201
    )

    def _run_test(self, params, expected):
        with self.assertRaises(expected):
            FileRule(*params)

    def test_missing_params_1(self):
        with self.assertRaises(TypeError):
            FileRule('/foo')

    def test_missing_params_2(self):
        with self.assertRaises(TypeError):
            FileRule('/foo', 'rw')

    def test_missing_params_3(self):
        with self.assertRaises(TypeError):
            FileRule('/foo', 'rw', 'ix')

    def test_missing_params_4(self):
        with self.assertRaises(TypeError):
            FileRule('/foo', 'rw', 'ix', '/bar')

    def test_deny_ix(self):
        with self.assertRaises(AppArmorException):
            FileRule('/foo', 'rw', 'ix', '/bar', False, False, False, deny=True)


class InvalidFileTest(AATest):
    def _check_invalid_rawrule(self, rawrule):
        obj = None
        self.assertFalse(FileRule.match(rawrule))
        with self.assertRaises(AppArmorException):
            obj = FileRule.create_instance(rawrule)

        self.assertIsNone(obj, 'FileRule handed back an object unexpectedly')

    def test_invalid_file_missing_comma_1(self):
        self._check_invalid_rawrule('file')  # missing comma

    def test_invalid_non_FileRule(self):
        self._check_invalid_rawrule('signal,')  # not a file rule


class BrokenFileTest(AATest):
    def AASetup(self):
        #                   path,   perms, exec_perms, target, owner, file_keyword, leading_perms
        self.obj = FileRule('/foo', 'rw',  'ix',       '/bar', False, False,        False)

    def test_empty_data_1(self):
        self.obj.path = ''
        # no path set, and ALL not set
        with self.assertRaises(AppArmorBug):
            self.obj.get_clean(1)

    def test_empty_data_2(self):
        self.obj.perms = ''
        self.obj.exec_perms = ''
        # no perms or exec_perms set, and ALL not set
        with self.assertRaises(AppArmorBug):
            self.obj.get_clean(1)

    def test_empty_data_3(self):
        self.obj.target = ''
        # no target set, and ALL not set
        with self.assertRaises(AppArmorBug):
            self.obj.get_clean(1)

    def test_unexpected_all_1(self):
        self.obj.all_paths = FileRule.ALL
        # all_paths and all_perms must be in sync
        with self.assertRaises(AppArmorBug):
            self.obj.get_clean(1)

    def test_unexpected_all_2(self):
        self.obj.all_perms = FileRule.ALL
        # all_paths and all_perms must be in sync
        with self.assertRaises(AppArmorBug):
            self.obj.get_clean(1)


class FileGlobTest(AATest):
    def _run_test(self, params, expected):
        exp_can_glob, exp_can_glob_ext, exp_rule_glob, exp_rule_glob_ext = expected

        # test glob()
        rule_obj = FileRule.create_instance(params)
        self.assertEqual(exp_can_glob, rule_obj.can_glob)
        self.assertEqual(exp_can_glob_ext, rule_obj.can_glob_ext)

        rule_obj.glob()
        self.assertEqual(rule_obj.get_clean(), exp_rule_glob)

        # test glob_ext()
        rule_obj = FileRule.create_instance(params)
        self.assertEqual(exp_can_glob, rule_obj.can_glob)
        self.assertEqual(exp_can_glob_ext, rule_obj.can_glob_ext)

        rule_obj.glob_ext()
        self.assertEqual(rule_obj.get_clean(), exp_rule_glob_ext)

    # These tests are meant to ensure AARE integration in FileRule works as expected.
    # test-aare.py has more comprehensive globbing tests.
    tests = (
        # rule               can glob  can glob_ext  globbed rule        globbed_ext rule
        ('/foo/bar r,',      (True,    True,         '/foo/* r,',        '/foo/bar r,')),
        ('/foo/* r,',        (True,    True,         '/** r,',           '/foo/* r,')),
        ('/foo/bar.xy r,',   (True,    True,         '/foo/* r,',        '/foo/*.xy r,')),
        ('/foo/*.xy r,',     (True,    True,         '/foo/* r,',        '/**.xy r,')),
        ('file,',            (False,   False,        'file,',            'file,')),  # bare 'file,' rules can't be globbed
        ('link /a/b -> /c,', (True,    True,         'link /a/* -> /c,', 'link /a/b -> /c,')),
    )


class WriteFileTest(AATest):
    def _run_test(self, rawrule, expected):
        self.assertTrue(FileRule.match(rawrule), 'FileRule.match() failed')
        obj = FileRule.create_instance(rawrule)
        clean = obj.get_clean()
        raw = obj.get_raw()

        self.assertEqual(expected.strip(), clean, 'unexpected clean rule')
        self.assertEqual(rawrule.strip(), raw, 'unexpected raw rule')

    tests = (
        #  raw rule                                      clean rule
        ('file,',                                        'file,'),
        ('              file        ,  # foo        ',   'file, # foo'),
        ('    audit     file /foo r,',                   'audit file /foo r,'),
        ('    audit     file /foo  lwr,',                'audit file /foo rwl,'),
        ('    audit     file /foo Pxrm -> bar,',         'audit file /foo mrPx -> bar,'),
        ('    deny      file /foo r,',                   'deny file /foo r,'),
        ('    deny      file /foo  wr,',                 'deny file /foo rw,'),
        ('    allow     file /foo Pxrm -> bar,',         'allow file /foo mrPx -> bar,'),
        ('   deny owner file /foo r,',                   'deny owner file /foo r,'),
        ('   deny owner file /foo  wr,',                 'deny owner file /foo rw,'),
        ('  allow owner file /foo Pxrm -> bar,',         'allow owner file /foo mrPx -> bar,'),
        ('    deny    owner  /foo r,',                   'deny owner /foo r,'),
        ('    deny    owner  /foo  wr,',                 'deny owner /foo rw,'),
        ('    allow   owner  /foo Pxrm -> bar,',         'allow owner /foo mrPx -> bar,'),
        ('                   /foo r,',                   '/foo r,'),
        ('                   /foo  lwr,',                '/foo rwl,'),
        ('                   /foo Pxrm -> bar,',         '/foo mrPx -> bar,'),
        ('  priority=-1      /foo Pxrm -> bar,',         'priority=-1 /foo mrPx -> bar,'),
        ('  priority=0       /foo Pxrm -> bar,',         'priority=0 /foo mrPx -> bar,'),
        ('  priority=343     /foo Pxrm -> bar,',         'priority=343 /foo mrPx -> bar,'),
        ('  priority=+65     /foo Pxrm -> bar,',         'priority=65 /foo mrPx -> bar,'),

        # with leading permissions
        ('    audit     file r      /foo,',              'audit file r /foo,'),
        ('    audit     file lwr    /foo,',              'audit file rwl /foo,'),
        ('    audit     file Pxrm   /foo -> bar,',       'audit file mrPx /foo -> bar,'),
        ('    deny      file r      /foo,',              'deny file r /foo,'),
        ('    deny      file wr     /foo  ,',            'deny file rw /foo,'),
        ('    allow     file Pxmr   /foo -> bar,',       'allow file mrPx /foo -> bar,'),
        ('   deny owner file r      /foo ,',             'deny owner file r /foo,'),
        ('   deny owner file wr     /foo  ,',            'deny owner file rw /foo,'),
        ('  allow owner file Pxrm   /foo -> bar,',       'allow owner file mrPx /foo -> bar,'),
        ('    deny    owner  r      /foo ,',             'deny owner r /foo,'),
        ('    deny    owner  wr     /foo  ,',            'deny owner rw /foo,'),
        ('    allow   owner  Pxrm   /foo -> bar,',       'allow owner mrPx /foo -> bar,'),
        ('                   r      /foo ,',             'r /foo,'),
        ('                   klwr   /foo  ,',            'rwlk /foo,'),
        ('                   Pxrm   /foo -> bar,',       'mrPx /foo -> bar,'),
        ('    priority=1     Pxrm   /foo -> bar,',       'priority=1 mrPx /foo -> bar,'),
        ('    priority=0     Pxrm   /foo -> bar,',       'priority=0 mrPx /foo -> bar,'),
        ('    priority=-22   Pxrm   /foo -> bar,',       'priority=-22 mrPx /foo -> bar,'),
        ('    priority=+78   Pxrm   /foo -> bar,',       'priority=78 mrPx /foo -> bar,'),

        # link rules
        ('  link    /foo    ->  /bar,',                  'link /foo -> /bar,'),
        ('  audit deny owner link subset /foo -> /bar,', 'audit deny owner link subset /foo -> /bar,'),
        ('                   link subset /foo -> /bar,', 'link subset /foo -> /bar,'),
        ('  priority=0  link    /foo    ->  /bar,',       'priority=0 link /foo -> /bar,'),
        ('  priority=12 link    /foo    ->  /bar,',       'priority=12 link /foo -> /bar,'),
        ('  priority=-1 link    /foo    ->  /bar,',       'priority=-1 link /foo -> /bar,'),
        ('  priority=+4 link    /foo    ->  /bar,',       'priority=4 link /foo -> /bar,'),
    )

    def test_write_manually_1(self):
        #              path,   perms, exec_perms, target, owner, file_keyword, leading_perms
        obj = FileRule('/foo', 'rw',  'Px',       '/bar', False, True,         False,       allow_keyword=True)

        expected = '    allow file /foo rwPx -> /bar,'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')

    def test_write_manually_2(self):
        #              path,   perms, exec_perms, target,       owner, file_keyword, leading_perms
        obj = FileRule('/foo', 'rw',  'x',        FileRule.ALL, True,  False,        True,        deny=True)

        expected = '    deny owner rwx /foo,'

        self.assertEqual(expected, obj.get_clean(2), 'unexpected clean rule')
        self.assertEqual(expected, obj.get_raw(2), 'unexpected raw rule')

    def test_write_any_exec(self):
        obj = FileRule('/foo', 'rw', FileRule.ANY_EXEC, '/bar', False, False, False)
        with self.assertRaises(AppArmorBug):
            obj.get_clean()


class FileCoveredTest(AATest):
    def _run_test(self, param, expected):
        obj = FileRule.create_instance(self.rule)
        check_obj = FileRule.create_instance(param)

        self.assertTrue(FileRule.match(param))

        self.assertEqual(obj.is_equal(check_obj), expected[0], 'Mismatch in is_equal, expected {}'.format(expected[0]))
        self.assertEqual(obj.is_equal(check_obj, True), expected[1], 'Mismatch in is_equal/strict, expected {}'.format(expected[1]))

        self.assertEqual(obj.is_covered(check_obj), expected[2], 'Mismatch in is_covered, expected {}'.format(expected[2]))
        self.assertEqual(obj.is_covered(check_obj, True, True), expected[3], 'Mismatch in is_covered/exact, expected {}'.format(expected[3]))


class FileCoveredTest_01(FileCoveredTest):
    rule = 'file /foo r,'
    tests = (
        #  rule                      equal  strict equal  covered  covered exact
        ('file /foo r,',            (True,  True,         True,    True)),
        ('file /foo r ,',           (True,  False,        True,    True)),
        ('allow file /foo r,',      (True,  False,        True,    True)),
        ('allow /foo r, # comment', (True,  False,        True,    True)),
        ('allow owner /foo r,',     (False, False,        True,    True)),
        ('/foo r -> bar,',          (False, False,        True,    True)),
        ('file r /foo,',            (True,  False,        True,    True)),
        ('allow file r /foo,',      (True,  False,        True,    True)),
        ('allow r /foo, # comment', (True,  False,        True,    True)),
        ('allow owner r /foo,',     (False, False,        True,    True)),
        ('r /foo -> bar,',          (False, False,        True,    True)),
        ('file,',                   (False, False,        False,   False)),
        ('file /foo w,',            (False, False,        False,   False)),
        ('file /foo rw,',           (False, False,        False,   False)),
        ('file /bar r,',            (False, False,        False,   False)),
        ('audit /foo r,',           (False, False,        False,   False)),
        ('audit file,',             (False, False,        False,   False)),
        ('audit deny /foo r,',      (False, False,        False,   False)),
        ('deny file /foo r,',       (False, False,        False,   False)),
        ('/foo rPx,',               (False, False,        False,   False)),
        ('/foo Pxr,',               (False, False,        False,   False)),
        ('/foo Px,',                (False, False,        False,   False)),
        ('/foo ix,',                (False, False,        False,   False)),
        ('/foo ix -> bar,',         (False, False,        False,   False)),
        ('/foo rPx -> bar,',        (False, False,        False,   False)),
    )


class FileCoveredTest_02(FileCoveredTest):
    rule = 'audit /foo r,'

    tests = (
        #  rule                      equal  strict equal  covered  covered exact
        ('file /foo r,',            (False, False,        True,    False)),
        ('allow file /foo r,',      (False, False,        True,    False)),
        ('allow /foo r, # comment', (False, False,        True,    False)),
        ('allow owner /foo r,',     (False, False,        True,    False)),
        ('/foo r -> bar,',          (False, False,        True,    False)),
        ('file r /foo,',            (False, False,        True,    False)),
        ('allow file r /foo,',      (False, False,        True,    False)),
        ('allow r /foo, # comment', (False, False,        True,    False)),
        ('allow owner r /foo,',     (False, False,        True,    False)),
        ('r /foo -> bar,',          (False, False,        True,    False)),  # XXX exact
        ('file,',                   (False, False,        False,   False)),
        ('file /foo w,',            (False, False,        False,   False)),
        ('file /foo rw,',           (False, False,        False,   False)),
        ('file /bar r,',            (False, False,        False,   False)),
        ('audit /foo r,',           (True,  True,         True,    True)),
        ('audit file,',             (False, False,        False,   False)),
        ('audit deny /foo r,',      (False, False,        False,   False)),
        ('deny file /foo r,',       (False, False,        False,   False)),
        ('/foo rPx,',               (False, False,        False,   False)),
        ('/foo Pxr,',               (False, False,        False,   False)),
        ('/foo Px,',                (False, False,        False,   False)),
        ('/foo ix,',                (False, False,        False,   False)),
        ('/foo ix -> bar,',         (False, False,        False,   False)),
        ('/foo rPx -> bar,',        (False, False,        False,   False)),
    )


class FileCoveredTest_03(FileCoveredTest):
    rule = '/foo mrwPx,'

    tests = (
        #  rule                      equal  strict equal  covered  covered exact
        ('file /foo r,',            (False, False,        True,    True)),
        ('allow file /foo r,',      (False, False,        True,    True)),
        ('allow /foo r, # comment', (False, False,        True,    True)),
        ('allow owner /foo r,',     (False, False,        True,    True)),
        ('/foo r -> bar,',          (False, False,        True,    True)),
        ('file r /foo,',            (False, False,        True,    True)),
        ('allow file r /foo,',      (False, False,        True,    True)),
        ('allow r /foo, # comment', (False, False,        True,    True)),
        ('allow owner r /foo,',     (False, False,        True,    True)),
        ('r /foo -> bar,',          (False, False,        True,    True)),
        ('file,',                   (False, False,        False,   False)),
        ('file /foo w,',            (False, False,        True,    True)),
        ('file /foo rw,',           (False, False,        True,    True)),
        ('file /bar r,',            (False, False,        False,   False)),
        ('audit /foo r,',           (False, False,        False,   False)),
        ('audit file,',             (False, False,        False,   False)),
        ('audit deny /foo r,',      (False, False,        False,   False)),
        ('deny file /foo r,',       (False, False,        False,   False)),
        ('/foo mrwPx,',             (True,  True,         True,    True)),
        ('/foo wPxrm,',             (True,  False,        True,    True)),
        ('/foo rm,',                (False, False,        True,    True)),
        ('/foo Px,',                (False, False,        True,    True)),
        ('/foo ix,',                (False, False,        False,   False)),
        ('/foo ix -> bar,',         (False, False,        False,   False)),
        ('/foo mrwPx -> bar,',      (False, False,        False,   False)),
    )


class FileCoveredTest_04(FileCoveredTest):
    rule = '/foo mrwPx -> bar,'

    tests = (
        #  rule                      equal  strict equal  covered  covered exact
        ('file /foo r,',            (False, False,        True,    True)),
        ('allow file /foo r,',      (False, False,        True,    True)),
        ('allow /foo r, # comment', (False, False,        True,    True)),
        ('allow owner /foo r,',     (False, False,        True,    True)),
        ('/foo r -> bar,',          (False, False,        True,    True)),
        ('file r /foo,',            (False, False,        True,    True)),
        ('allow file r /foo,',      (False, False,        True,    True)),
        ('allow r /foo, # comment', (False, False,        True,    True)),
        ('allow owner r /foo,',     (False, False,        True,    True)),
        ('r /foo -> bar,',          (False, False,        True,    True)),
        ('file,',                   (False, False,        False,   False)),
        ('file /foo w,',            (False, False,        True,    True)),
        ('file /foo rw,',           (False, False,        True,    True)),
        ('file /bar r,',            (False, False,        False,   False)),
        ('audit /foo r,',           (False, False,        False,   False)),
        ('audit file,',             (False, False,        False,   False)),
        ('audit deny /foo r,',      (False, False,        False,   False)),
        ('deny file /foo r,',       (False, False,        False,   False)),
        ('/foo mrwPx,',             (False, False,        False,   False)),
        ('/foo wPxrm,',             (False, False,        False,   False)),
        ('/foo rm,',                (False, False,        True,    True)),
        ('/foo Px,',                (False, False,        False,   False)),
        ('/foo ix,',                (False, False,        False,   False)),
        ('/foo ix -> bar,',         (False, False,        False,   False)),
        ('/foo mrwPx -> bar,',      (True,  True,         True,    True)),
    )


class FileCoveredTest_05(FileCoveredTest):
    rule = 'file,'

    tests = (
        #   rule                     equal  strict equal  covered  covered exact
        ('file /foo r,',            (False, False,        True,    True)),
        ('allow file /foo r,',      (False, False,        True,    True)),
        ('allow /foo r, # comment', (False, False,        True,    True)),
        ('allow owner /foo r,',     (False, False,        True,    True)),
        ('/foo r -> bar,',          (False, False,        True,    True)),
        ('file r /foo,',            (False, False,        True,    True)),
        ('allow file r /foo,',      (False, False,        True,    True)),
        ('allow r /foo, # comment', (False, False,        True,    True)),
        ('allow owner r /foo,',     (False, False,        True,    True)),
        ('r /foo -> bar,',          (False, False,        True,    True)),
        ('file,',                   (True,  True,         True,    True)),
        ('file /foo w,',            (False, False,        True,    True)),
        ('file /foo rw,',           (False, False,        True,    True)),
        ('file /bar r,',            (False, False,        True,    True)),
        ('audit /foo r,',           (False, False,        False,   False)),
        ('audit file,',             (False, False,        False,   False)),
        ('audit deny /foo r,',      (False, False,        False,   False)),
        ('deny file /foo r,',       (False, False,        False,   False)),
        ('/foo mrwPx,',             (False, False,        False,   False)),
        ('/foo wPxrm,',             (False, False,        False,   False)),
        ('/foo rm,',                (False, False,        True,    True)),
        ('/foo Px,',                (False, False,        False,   False)),
        ('/foo ix,',                (False, False,        False,   False)),
        ('/foo ix -> bar,',         (False, False,        False,   False)),
        ('/foo mrwPx -> bar,',      (False, False,        False,   False)),
    )


class FileCoveredTest_06(FileCoveredTest):
    rule = 'deny /foo w,'

    tests = (
        #   rule          equal  strict equal  covered  covered exact
        ('/foo w,',      (False, False,        False,   False)),
        ('/foo a,',      (False, False,        False,   False)),
        ('deny /foo w,', (True,  True,         True,    True)),
        ('deny /foo a,', (False, False,        True,    True)),
    )


class FileCoveredTest_07(FileCoveredTest):
    rule = '/foo w,'

    tests = (
        #   rule          equal  strict equal  covered  covered exact
        ('/foo w,',      (True,  True,         True,    True)),
        ('/foo a,',      (False, False,        True,    True)),
        ('deny /foo w,', (False, False,        False,   False)),
        ('deny /foo a,', (False, False,        False,   False)),
    )


class FileCoveredTest_08(FileCoveredTest):
    rule = 'link /foo -> /bar,'

    tests = (
        #   rule                        equal  strict equal  covered  covered exact
        ('link /foo -> /bar,',        (True,  True,          True,    True)),
        ('link /asdf -> /bar,',       (False, False,         False,   False)),
        ('link /foo -> /asdf,',       (False, False,         False,   False)),
        ('deny link /foo -> /bar,',   (False, False,         False,   False)),
        ('deny link /foo -> /bar,',   (False, False,         False,   False)),
        ('link subset /foo -> /bar,', (False, False,         True,    True)),  # subset makes the rule more strict
        # ('/foo l -> /bar,',           (?,     ?,             ?,       ?)),  # TODO
        # ('l /foo -> /bar,',           (?,     ?,             ?,       ?)),  # TODO
    )


class FileCoveredTest_09(FileCoveredTest):
    rule = 'link subset /foo -> /bar,'
    tests = (
        #   rule                            equal  strict equal  covered  covered exact
        ('link subset /foo -> /bar,',      (True,  True,         True,    True)),
        ('link subset /asdf -> /bar,',     (False, False,        False,   False)),
        ('link subset /foo -> /asdf,',     (False, False,        False,   False)),
        ('deny link subset /foo -> /bar,', (False, False,        False,   False)),
        ('deny link subset /foo -> /bar,', (False, False,        False,   False)),
        ('link /foo -> /bar,',             (False, False,        False,   False)),  # no subset means more permissions
        # ('/foo l -> /bar,',                (?,     ?,            ?,       ?)),  # TODO
        # ('l /foo -> /bar,',                (?,     ?,            ?,       ?)),  # TODO
    )


class FileCoveredTest_ManualOrInvalid(AATest):
    def AASetup(self):
        #                   path,   perms, exec_perms, target, owner, file_keyword, leading_perms
        self.obj = FileRule('/foo', 'rw',  'ix',       '/bar', False, False,        False)
        self.testobj = FileRule('/foo', 'rw', 'ix', '/bar', False, False, False)

    def test_covered_owner_1(self):
        # testobj with 'owner'
        self.testobj = FileRule('/foo', 'rw', 'ix', '/bar', True, False, False)
        self.assertTrue(self.obj.is_covered(self.testobj))

    def test_covered_owner_2(self):
        # obj with 'owner'
        self.obj = FileRule('/foo', 'rw', 'ix', '/bar', True, False, False)
        self.assertFalse(self.obj.is_covered(self.testobj))

    def test_equal_all_perms(self):
        self.testobj.all_perms = True  # that makes testobj invalid, but that's the only way to survive the 'perms' comparison
        self.assertFalse(self.obj.is_equal(self.testobj))

    def test_equal_file_keyword(self):
        # testobj with file_keyword
        self.testobj = FileRule('/foo', 'rw', 'ix', '/bar', False, True, False)
        self.assertTrue(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_equal_file_leading_perms(self):
        # testobj with leading_perms
        self.testobj = FileRule('/foo', 'rw', 'ix', '/bar', False, False, True)
        self.assertTrue(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_covered_anyperm_1(self):
        self.obj = FileRule('/foo', 'rw', None, '/bar', False, False, False)
        self.testobj = FileRule('/foo', 'rw', FileRule.ANY_EXEC, '/bar', False, False, False)
        self.assertFalse(self.obj.is_covered(self.testobj))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_covered_anyperm_2(self):
        self.testobj = FileRule('/foo', 'rw', FileRule.ANY_EXEC, '/bar', False, False, False)
        self.assertTrue(self.obj.is_covered(self.testobj))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_covered_anyperm_3(self):
        # make sure a different exec target gets ignored with ANY_EXEC
        self.testobj = FileRule('/foo', 'rw', FileRule.ANY_EXEC, '/xyz', False, False, False)
        self.assertTrue(self.obj.is_covered(self.testobj))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_covered_anyperm_4(self):
        # make sure a different exec target gets ignored with ANY_EXEC
        self.testobj = FileRule('/foo', 'rw', FileRule.ANY_EXEC, FileRule.ALL, False, False, False)
        self.assertTrue(self.obj.is_covered(self.testobj))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_covered_anyperm_5(self):
        # even with ANY_EXEC, a different link target causes a mismatch
        self.testobj = FileRule('/foo', 'rwl', FileRule.ANY_EXEC, '/xyz', False, False, False)
        self.assertFalse(self.obj.is_covered(self.testobj))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_covered_anyperm_6(self):
        # even with ANY_EXEC, a different link target causes a mismatch
        self.testobj = FileRule('/foo', 'rwl', FileRule.ANY_EXEC, FileRule.ALL, False, False, False)
        self.assertFalse(self.obj.is_covered(self.testobj))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_covered_anyperm_7(self):
        self.obj = FileRule('/foo', 'rw', 'x', '/bar', False, False, False, deny=True)
        self.testobj = FileRule('/foo', 'rw', FileRule.ANY_EXEC, '/bar', False, False, False)
        self.assertFalse(self.obj.is_covered(self.testobj))
        self.assertTrue(self.obj.is_covered(self.testobj, check_allow_deny=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=False))
        self.assertFalse(self.obj.is_equal(self.testobj, strict=True))

    def test_borked_obj_is_covered_1(self):
        self.testobj.path = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_2(self):
        self.testobj.perms = set()
        self.testobj.exec_perms = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_borked_obj_is_covered_3(self):
        self.testobj.target = ''

        with self.assertRaises(AppArmorBug):
            self.obj.is_covered(self.testobj)

    def test_invalid_is_covered(self):
        raw_rule = 'file,'

        class SomeOtherClass(FileRule):
            pass

        obj = FileRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_covered(testobj)

    def test_invalid_is_equal(self):
        raw_rule = 'file,'

        class SomeOtherClass(FileRule):
            pass

        obj = FileRule.create_instance(raw_rule)
        testobj = SomeOtherClass.create_instance(raw_rule)  # different type
        with self.assertRaises(AppArmorBug):
            obj.is_equal(testobj)


class FileSeverityTest(AATest):
    tests = (
        ('/usr/bin/whatis ix,',  5),
        ('/etc ix,',             'unknown'),
        ('/dev/doublehit ix,',   0),
        ('/dev/doublehit rix,',  4),
        ('/dev/doublehit rwix,', 8),
        ('/dev/tty10 rwix,',     9),
        ('/var/adm/foo/** rix,', 3),
        ('/etc/apparmor/** r,',  6),
        ('/etc/** r,',           'unknown'),
        ('/usr/foo@bar r,',      'unknown'),  # filename containing @
        ('/home/foo@bar rw,',    6),  # filename containing @
        ('file,',                'unknown'),  # bare file rule XXX should return maximum severity
    )

    def _run_test(self, params, expected):
        sev_db = severity.Severity('../severity.db', 'unknown')
        obj = FileRule.create_instance(params)
        rank = obj.severity(sev_db)
        self.assertEqual(rank, expected)


class FileLogprofHeaderTest(AATest):
    tests = (
        # log event                        old perms ALL / owner
        (('file,',                             set(),    set()),     [                                     _('Path'), _('ALL'),                                _('New Mode'), _('ALL')]),  # noqa: E201
        (('/foo r,',                           set(),    set()),     [                                     _('Path'), '/foo',                                  _('New Mode'), 'r']),  # noqa: E201
        (('file /bar Px -> foo,',              set(),    set()),     [                                     _('Path'), '/bar',                                  _('New Mode'), 'Px -> foo']),  # noqa: E201
        (('deny file,',                        set(),    set()),     [_('Qualifier'), 'deny',              _('Path'), _('ALL'),                                _('New Mode'), _('ALL')]),
        (('allow file /baz rwk,',              set(),    set()),     [_('Qualifier'), 'allow',             _('Path'), '/baz',                                  _('New Mode'), 'rwk']),
        (('audit file /foo mr,',               set(),    set()),     [_('Qualifier'), 'audit',             _('Path'), '/foo',                                  _('New Mode'), 'mr']),
        (('audit deny /foo wk,',               set(),    set()),     [_('Qualifier'), 'audit deny',        _('Path'), '/foo',                                  _('New Mode'), 'wk']),
        (('priority=1 file,',                  set(),    set()),     [_('Qualifier'), 'priority=1',        _('Path'), _('ALL'),                                _('New Mode'), _('ALL')]),
        (('priority=-1 audit file /foo mr,',   set(),    set()),     [_('Qualifier'), 'priority=-1 audit', _('Path'), '/foo',                                  _('New Mode'), 'mr']),
        (('owner file /foo ix,',               set(),    set()),     [                                     _('Path'), '/foo',                                  _('New Mode'), 'owner ix']),  # noqa: E201
        (('audit deny file /foo rlx -> /baz,', set(),    set()),     [_('Qualifier'), 'audit deny',        _('Path'), '/foo',                                  _('New Mode'), 'rlx -> /baz']),
        (('/foo rw,',                          set('r'), set()),     [                                     _('Path'), '/foo', _('Old Mode'), _('r'),           _('New Mode'), 'rw']),  # noqa: E201
        (('/foo rw,',                          set(),    set('rw')), [                                     _('Path'), '/foo', _('Old Mode'), _('owner rw'),    _('New Mode'), 'rw']),  # noqa: E201
        (('/foo mrw,',                         set('r'), set('k')),  [                                     _('Path'), '/foo', _('Old Mode'), _('r + owner k'), _('New Mode'), 'mrw']),  # noqa: E201
        (('/foo mrw,',                         set('r'), set('rk')), [                                     _('Path'), '/foo', _('Old Mode'), _('r + owner k'), _('New Mode'), 'mrw']),  # noqa: E201
        (('link /foo -> /bar,',                set(),    set()),     [                                     _('Path'), '/foo',                                  _('New Mode'), 'link -> /bar']),  # noqa: E201
        (('link subset /foo -> /bar,',         set(),    set()),     [                                     _('Path'), '/foo',                                  _('New Mode'), 'link subset -> /bar']),  # noqa: E201
    )

    def _run_test(self, params, expected):
        obj = FileRule.create_instance(params[0])
        if params[1] or params[2]:
            obj.original_perms = {'allow': {'all': params[1], 'owner': params[2]}}
        self.assertEqual(obj.logprof_header(), expected)

    def test_empty_original_perms(self):
        obj = FileRule.create_instance('/foo rw,')
        obj.original_perms = {'allow': {'all': set(), 'owner': set()}}
        self.assertEqual(obj.logprof_header(), [_('Path'), '/foo', _('New Mode'), 'rw'])

    def test_implicit_original_perms(self):
        obj = FileRule.create_instance('/foo rw,')
        obj.original_perms = {'allow': {'all': FileRule.ALL, 'owner': set()}}
        self.assertEqual(obj.logprof_header(), [_('Path'), '/foo', _('Old Mode'), _('mrwlkix'), _('New Mode'), 'rw'])

    def test_owner_implicit_original_perms(self):
        obj = FileRule.create_instance('/foo rw,')
        obj.original_perms = {'allow': {'all': set(), 'owner': FileRule.ALL}}
        self.assertEqual(obj.logprof_header(), [_('Path'), '/foo', _('Old Mode'), _('owner mrwlkix'), _('New Mode'), 'rw'])


class FileEditHeaderTest(AATest):
    def _run_test(self, params, expected):
        rule_obj = FileRule.create_instance(params)
        self.assertEqual(rule_obj.can_edit, True)
        prompt, path_to_edit = rule_obj.edit_header()
        self.assertEqual(path_to_edit, expected)

    tests = (
        ('/foo/bar/baz r,',       '/foo/bar/baz'),
        ('/foo/**/baz r,',        '/foo/**/baz'),
        ('link /foo/** -> /bar,', '/foo/**'),
    )

    def test_edit_header_bare_file(self):
        rule_obj = FileRule.create_instance('file,')
        self.assertEqual(rule_obj.can_edit, False)
        with self.assertRaises(AppArmorBug):
            rule_obj.edit_header()


class FileValidateAndStoreEditTest(AATest):
    def _run_test(self, params, expected):
        rule_obj = FileRule('/foo/bar/baz', 'r', None, FileRule.ALL, False, False, False, log_event=True)

        self.assertEqual(rule_obj.validate_edit(params), expected)

        rule_obj.store_edit(params)
        self.assertEqual(rule_obj.get_raw(), params + ' r,')

    tests = (
        # edited path     match
        ('/foo/bar/baz',  True),
        ('/foo/bar/*',    True),
        ('/foo/bar/???',  True),
        ('/foo/xy**',     False),
        ('/foo/bar/baz/', False),
    )

    def test_validate_not_a_path(self):
        rule_obj = FileRule.create_instance('/foo/bar/baz r,')

        with self.assertRaises(AppArmorException):
            rule_obj.validate_edit('foo/bar/baz')

        with self.assertRaises(AppArmorException):
            rule_obj.store_edit('foo/bar/baz')

    def test_validate_edit_bare_file(self):
        rule_obj = FileRule.create_instance('file,')
        self.assertEqual(rule_obj.can_edit, False)

        with self.assertRaises(AppArmorBug):
            rule_obj.validate_edit('/foo/bar')

        with self.assertRaises(AppArmorBug):
            rule_obj.store_edit('/foo/bar')


# --- tests for FileRuleset --- #

class FileRulesTest(AATest):
    def test_empty_ruleset(self):
        ruleset = FileRuleset()
        ruleset_2 = FileRuleset()
        self.assertEqual([], ruleset.get_raw(2))
        self.assertEqual([], ruleset.get_clean(2))
        self.assertEqual([], ruleset_2.get_raw(2))
        self.assertEqual([], ruleset_2.get_clean(2))

    def test_ruleset_1(self):
        ruleset = FileRuleset()
        rules = [
            '         file             ,        ',
            '   file /foo rw,',
            '  file /bar r,',
        ]

        expected_raw = [
            'file             ,',
            'file /foo rw,',
            'file /bar r,',
            '',
        ]

        expected_clean = [
            'file /bar r,',
            'file /foo rw,',
            'file,',
            '',
        ]

        deleted = 0
        for rule in rules:
            deleted += ruleset.add(FileRule.create_instance(rule))

        self.assertEqual(deleted, 0)
        self.assertEqual(expected_raw, ruleset.get_raw())
        self.assertEqual(expected_clean, ruleset.get_clean())

    def test_ruleset_2(self):
        ruleset = FileRuleset()
        rules = [
            '/foo Px,',
            '/bar    Cx    ->     bar_child ,',
            'deny /asdf w,',
        ]

        expected_raw = [
            '  /foo Px,',
            '  /bar    Cx    ->     bar_child ,',
            '  deny /asdf w,',
            '',
        ]

        expected_clean = [
            '  deny /asdf w,',
            '',
            '  /bar Cx -> bar_child,',
            '  /foo Px,',
            '',
        ]

        deleted = 0
        for rule in rules:
            deleted += ruleset.add(FileRule.create_instance(rule))

        self.assertEqual(deleted, 0)
        self.assertEqual(expected_raw, ruleset.get_raw(1))
        self.assertEqual(expected_clean, ruleset.get_clean(1))

    def test_ruleset_cleanup_add_1(self):
        ruleset = FileRuleset()
        rules = [
            '/foo/bar r,',
            '/foo/baz rw,',
            '/foo/baz rwk,',
        ]

        rules_with_cleanup = [
            '/foo/* r,',
        ]

        expected_raw = [
            '  /foo/baz rw,',
            '  /foo/baz rwk,',
            '  /foo/* r,',
            '',
        ]

        expected_clean = [
            '  /foo/* r,',
            '  /foo/baz rw,',
            '  /foo/baz rwk,',
            '',
        ]

        deleted = 0
        for rule in rules:
            deleted += ruleset.add(FileRule.create_instance(rule))

        self.assertEqual(deleted, 0)  # rules[] are added without cleanup mode, so the superfluous '/foo/baz rw,' should be kept

        for rule in rules_with_cleanup:
            deleted += ruleset.add(FileRule.create_instance(rule), cleanup=True)

        self.assertEqual(deleted, 1)  # rules_with_cleanup made '/foo/bar r,' superfluous
        self.assertEqual(expected_raw, ruleset.get_raw(1))
        self.assertEqual(expected_clean, ruleset.get_clean(1))


# class FileDeleteTest(AATest):
#     pass


class FileGetRulesForPath(AATest):
    tests = (
        #  path                                 audit  deny    expected
        (('/etc/foo/dovecot.conf',              False, False), ['/etc/foo/* r,', '/etc/foo/dovecot.conf rw,',                                 '']),
        (('/etc/foo/foo.conf',                  False, False), ['/etc/foo/* r,',                                                              '']),
        (('/etc/foo/dovecot-database.conf.ext', False, False), ['/etc/foo/* r,', '/etc/foo/dovecot-database.conf.ext w,',                     '']),
        (('/etc/foo/auth.d/authfoo.conf',       False, False), ['/etc/foo/{auth,conf}.d/*.conf r,', '/etc/foo/{auth,conf}.d/authfoo.conf w,', '']),
        (('/etc/foo/dovecot-deny.conf',         False, False), ['deny /etc/foo/dovecot-deny.conf r,', '', '/etc/foo/* r,',                    '']),
        (('/foo/bar',                           False, True),  [                                                                                ]),  # noqa: E201,E202
        (('/etc/foo/dovecot-deny.conf',         False, True),  ['deny /etc/foo/dovecot-deny.conf r,',                                         '']),
        (('/etc/foo/foo.conf',                  False, True),  [                                                                                ]),  # noqa: E201,E202
        (('/etc/foo/owner.conf',                False, False), ['/etc/foo/* r,', 'owner /etc/foo/owner.conf w,',                              '']),
    )

    def _run_test(self, params, expected):
        rules = (
            '/etc/foo/* r,',
            '/etc/foo/dovecot.conf rw,',
            '/etc/foo/{auth,conf}.d/*.conf r,',
            '/etc/foo/{auth,conf}.d/authfoo.conf w,',
            '/etc/foo/dovecot-database.conf.ext w,',
            'owner /etc/foo/owner.conf w,',
            'deny /etc/foo/dovecot-deny.conf r,',
        )

        ruleset = FileRuleset()
        for rule in rules:
            ruleset.add(FileRule.create_instance(rule))

        matching = ruleset.get_rules_for_path(*params)
        self. assertEqual(matching.get_clean(), expected)


class FileGetPermsForPath_1(AATest):
    tests = (
        #  path                                 audit   deny    expected
        (('/etc/foo/dovecot.conf',              False,  False), {'allow': {'all': {'r', 'w'}, 'owner': set()}, 'deny': {'all': set(), 'owner': set()}, 'paths': {'/etc/foo/*', '/etc/foo/dovecot.conf'}}),
        (('/etc/foo/foo.conf',                  False,  False), {'allow': {'all': {'r'},      'owner': set()}, 'deny': {'all': set(), 'owner': set()}, 'paths': {'/etc/foo/*'}}),
        (('/etc/foo/dovecot-database.conf.ext', False,  False), {'allow': {'all': {'r', 'w'}, 'owner': set()}, 'deny': {'all': set(), 'owner': set()}, 'paths': {'/etc/foo/*', '/etc/foo/dovecot-database.conf.ext'}}),
        (('/etc/foo/auth.d/authfoo.conf',       False,  False), {'allow': {'all': {'r', 'w'}, 'owner': set()}, 'deny': {'all': set(), 'owner': set()}, 'paths': {'/etc/foo/{auth,conf}.d/*.conf', '/etc/foo/{auth,conf}.d/authfoo.conf'}}),
        (('/etc/foo/dovecot-deny.conf',         False,  False), {'allow': {'all': {'r'},      'owner': set()}, 'deny': {'all': {'r'}, 'owner': set()}, 'paths': {'/etc/foo/*', '/etc/foo/dovecot-deny.conf'}}),
        (('/foo/bar',                           False,  True),  {'allow': {'all': set(),      'owner': set()}, 'deny': {'all': set(), 'owner': set()}, 'paths': set()}),
        (('/etc/foo/dovecot-deny.conf',         False,  True),  {'allow': {'all': set(),      'owner': set()}, 'deny': {'all': {'r'}, 'owner': set()}, 'paths': {'/etc/foo/dovecot-deny.conf'}}),
        (('/etc/foo/foo.conf',                  False,  True),  {'allow': {'all': set(),      'owner': set()}, 'deny': {'all': set(), 'owner': set()}, 'paths': set()}),
        (('/usr/lib/dovecot/config',            False,  False), {'allow': {'all': set(),      'owner': set()}, 'deny': {'all': set(), 'owner': set()}, 'paths': set()}),  # exec perms are not honored by get_perms_for_path()
    )

    def _run_test(self, params, expected):
        rules = (
            '/etc/foo/* r,',
            '/etc/foo/dovecot.conf rw,',
            '/etc/foo/{auth,conf}.d/*.conf r,',
            '/etc/foo/{auth,conf}.d/authfoo.conf w,',
            '/etc/foo/dovecot-database.conf.ext w,',
            'deny /etc/foo/dovecot-deny.conf r,',
            '/usr/lib/dovecot/config ix,',
        )

        ruleset = FileRuleset()
        for rule in rules:
            ruleset.add(FileRule.create_instance(rule))

        perms = ruleset.get_perms_for_path(*params)
        self. assertEqual(perms, expected)


class FileGetPermsForPath_2(AATest):
    tests = (
        #  path                                 audit   deny    expected
        (('/etc/foo/dovecot.conf',              False,  False), {'allow': {'all': FileRule.ALL, 'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': {'/etc/foo/*', '/etc/foo/dovecot.conf'}}),
        (('/etc/foo/dovecot.conf',              True,   False), {'allow': {'all': {'r', 'w'},   'owner': set()}, 'deny': {'all': set(),        'owner': set()}, 'paths': {'/etc/foo/dovecot.conf'}}),
        (('/etc/foo/foo.conf',                  False,  False), {'allow': {'all': FileRule.ALL, 'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': {'/etc/foo/*'}}),
        (('/etc/foo/dovecot-database.conf.ext', False,  False), {'allow': {'all': FileRule.ALL, 'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': {'/etc/foo/*', '/etc/foo/dovecot-database.conf.ext'}}),
        (('/etc/foo/auth.d/authfoo.conf',       False,  False), {'allow': {'all': FileRule.ALL, 'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': {'/etc/foo/{auth,conf}.d/*.conf', '/etc/foo/{auth,conf}.d/authfoo.conf'}}),
        (('/etc/foo/auth.d/authfoo.conf',       True,   False), {'allow': {'all': {'w'},        'owner': set()}, 'deny': {'all': set(),        'owner': set()}, 'paths': {'/etc/foo/{auth,conf}.d/authfoo.conf'}}),
        (('/etc/foo/dovecot-deny.conf',         False,  False), {'allow': {'all': FileRule.ALL, 'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': {'/etc/foo/*', '/etc/foo/dovecot-deny.conf'}}),
        (('/foo/bar',                           False,  True),  {'allow': {'all': set(),        'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': set()}),
        (('/etc/foo/dovecot-deny.conf',         False,  True),  {'allow': {'all': set(),        'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': {'/etc/foo/dovecot-deny.conf'}}),
        (('/etc/foo/foo.conf',                  False,  True),  {'allow': {'all': set(),        'owner': set()}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': set()}),
        # (('/etc/foo/owner.conf',                False,  True),  {'allow': {'all': set(),        'owner': {'w'}}, 'deny': {'all': FileRule.ALL, 'owner': set()}, 'paths': {'/etc/foo/owner.conf'}}), # XXX doesn't work yet
    )

    def _run_test(self, params, expected):
        rules = (
            '/etc/foo/* r,',
            'audit /etc/foo/dovecot.conf rw,',
            '/etc/foo/{auth,conf}.d/*.conf r,',
            'audit /etc/foo/{auth,conf}.d/authfoo.conf w,',
            '/etc/foo/dovecot-database.conf.ext w,',
            'deny /etc/foo/dovecot-deny.conf r,',
            'file,',
            'owner /etc/foo/owner.conf w,',
            'deny file,',
        )

        ruleset = FileRuleset()
        for rule in rules:
            ruleset.add(FileRule.create_instance(rule))

        perms = ruleset.get_perms_for_path(*params)
        self. assertEqual(perms, expected)


class FileGetExecRulesForPath_1(AATest):
    tests = (
        ('/bin/foo', ['audit /bin/foo ix,', '']),
        ('/bin/bar', ['deny /bin/bar x,', '']),
        ('/foo', []),
    )

    def _run_test(self, params, expected):
        rules = (
            '/foo r,',
            'audit /bin/foo ix,',
            '/bin/b* Px,',
            'deny /bin/bar x,',
        )

        ruleset = FileRuleset()
        for rule in rules:
            ruleset.add(FileRule.create_instance(rule))

        perms = ruleset.get_exec_rules_for_path(params)
        matches = perms.get_clean()
        self. assertEqual(matches, expected)


class FileGetExecRulesForPath_2(AATest):
    tests = (
        ('/bin/foo', ['audit /bin/foo ix,', '']),
        ('/bin/bar', ['deny /bin/bar x,', '', '/bin/b* Px,', '']),
        ('/foo', []),
    )

    def _run_test(self, params, expected):
        rules = (
            '/foo r,',
            'audit /bin/foo ix,',
            '/bin/b* Px,',
            'deny /bin/bar x,',
        )

        ruleset = FileRuleset()
        for rule in rules:
            ruleset.add(FileRule.create_instance(rule))

        perms = ruleset.get_exec_rules_for_path(params, only_exact_matches=False)
        matches = perms.get_clean()
        self. assertEqual(matches, expected)


class FileGetExecConflictRules_1(AATest):
    tests = (
        ('/bin/foo ix,', ['/bin/foo Px,', '']),
        ('/bin/bar Px,', ['deny /bin/bar x,', '', '/bin/bar cx,', '']),
        ('/bin/bar cx,', ['deny /bin/bar x,', '']),
        ('/bin/foo r,', []),
    )

    def _run_test(self, params, expected):
        rules = (
            '/foo r,',
            'audit /bin/foo ix,',
            '/bin/foo Px,',
            '/bin/b* Px,',
            '/bin/bar cx,',
            'deny /bin/bar x,',
        )

        ruleset = FileRuleset()
        for rule in rules:
            ruleset.add(FileRule.create_instance(rule))

        rule_obj = FileRule.create_instance(params)
        conflicts = ruleset.get_exec_conflict_rules(rule_obj)
        self. assertEqual(conflicts.get_clean(), expected)


class FileCreateFromEvent(AATest):

    def test_exec_rule(self):
        parser = ReadLog('', '', '')
        raw_ev = '[258945.534540] audit: type=1400 audit(1725865139.443:401): apparmor="DENIED" operation="exec" class="file" profile="foo" name="/usr/bin/ls" pid=130888 comm="foo" requested_mask="x" denied_mask="x" fsuid=1000 ouid=0'
        hl = hasher()
        ev = parser.parse_event(raw_ev)
        FileRule.hashlog_from_event(hl, ev)

        expected = {'/usr/bin/ls': {'': True}}
        self.assertEqual(hl, expected)

        fr = FileRule.from_hashlog(hl)

        expected = FileRule('/usr/bin/ls', None, FileRule.ANY_EXEC, FileRule.ALL, False)

        self.assertTrue(expected.is_equal(next(fr)))
        with self.assertRaises(StopIteration):
            next(fr)

    def test_filemode_write_append(self):
        parser = ReadLog('', '', '')
        events = [
            '[ 9614.885136] audit: type=1400 audit(1720429924.397:191): apparmor="DENIED" operation="open" class="file" profile="/home/user/test/a" name="/home/user/test/foo" pid=24460 comm="a" requested_mask="w" denied_mask="w" fsuid=1000 ouid=1000',
            '[ 9614.885149] audit: type=1400 audit(1720429924.397:192): apparmor="DENIED" operation="open" class="file" profile="/home/user/test/a" name="/home/user/test/foo" pid=24460 comm="a" requested_mask="a" denied_mask="a" fsuid=1000 ouid=1000'
        ]
        hl = hasher()
        for raw_ev in events:
            ev = parser.parse_event(raw_ev)
            FileRule.hashlog_from_event(hl, ev)

        expected = {'/home/user/test/foo': {True: {'w': True, 'a': True}}}
        self.assertEqual(hl, expected)

        fr = FileRule.from_hashlog(hl)

        expected = FileRule('/home/user/test/foo', 'w', None, FileRule.ALL, True)

        self.assertTrue(expected.is_equal(next(fr)))
        with self.assertRaises(StopIteration):
            next(fr)


class FileInvalidCreateFromEvent(AATest):
    def test_write_append(self):
        parser = ReadLog('', '', '')
        raw_ev = '[258145.094974] audit: type=1400 audit(1725864339.004:396): apparmor="DENIED" operation="exec" class="file" profile="foo" name="" pid=125456 comm="foo" requested_mask="x" denied_mask="x" fsuid=1000 ouid=1000'  # Name is empty
        hl = hasher()
        ev = parser.parse_event(raw_ev)
        with self.assertRaises(AppArmorException):
            FileRule.hashlog_from_event(hl, ev)


setup_all_loops(__name__)
if __name__ == '__main__':
    unittest.main(verbosity=1)
